#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

from openstack.block_storage.v3 import snapshot as _snapshot
from openstack.block_storage.v3 import volume as _volume
from openstack import exceptions as sdk_exceptions
from openstack.identity.v3 import project as _project
from openstack.test import fakes as sdk_fakes
from osc_lib.cli import format_columns
from osc_lib import exceptions

from openstackclient.tests.unit import utils as test_utils
from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes
from openstackclient.volume.v3 import volume_snapshot


class TestVolumeSnapshotCreate(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.volume = sdk_fakes.generate_fake_resource(_volume.Volume)
        self.volume_client.find_volume.return_value = self.volume
        self.snapshot = sdk_fakes.generate_fake_resource(
            _snapshot.Snapshot, volume_id=self.volume.id
        )
        self.volume_client.create_snapshot.return_value = self.snapshot
        self.volume_client.manage_snapshot.return_value = self.snapshot

        self.columns = (
            'created_at',
            'description',
            'id',
            'name',
            'properties',
            'size',
            'status',
            'volume_id',
        )
        self.data = (
            self.snapshot.created_at,
            self.snapshot.description,
            self.snapshot.id,
            self.snapshot.name,
            format_columns.DictColumn(self.snapshot.metadata),
            self.snapshot.size,
            self.snapshot.status,
            self.snapshot.volume_id,
        )

        self.cmd = volume_snapshot.CreateVolumeSnapshot(self.app, None)

    def test_snapshot_create(self):
        arglist = [
            "--volume",
            self.snapshot.volume_id,
            "--description",
            self.snapshot.description,
            "--force",
            '--property',
            'Alpha=a',
            '--property',
            'Beta=b',
            self.snapshot.name,
        ]
        verifylist = [
            ("volume", self.snapshot.volume_id),
            ("description", self.snapshot.description),
            ("force", True),
            ('properties', {'Alpha': 'a', 'Beta': 'b'}),
            ("snapshot_name", self.snapshot.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.find_volume.assert_called_once_with(
            self.snapshot.volume_id, ignore_missing=False
        )
        self.volume_client.create_snapshot.assert_called_with(
            volume_id=self.snapshot.volume_id,
            force=True,
            name=self.snapshot.name,
            description=self.snapshot.description,
            metadata={'Alpha': 'a', 'Beta': 'b'},
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_snapshot_create_without_name(self):
        arglist = [
            "--volume",
            self.snapshot.volume_id,
        ]
        verifylist = [
            ("volume", self.snapshot.volume_id),
        ]
        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_snapshot_create_without_volume(self):
        arglist = [
            "--description",
            self.snapshot.description,
            "--force",
            self.snapshot.name,
        ]
        verifylist = [
            ("description", self.snapshot.description),
            ("force", True),
            ("snapshot_name", self.snapshot.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.find_volume.assert_called_once_with(
            self.snapshot.name, ignore_missing=False
        )
        self.volume_client.create_snapshot.assert_called_with(
            volume_id=self.snapshot.volume_id,
            force=True,
            name=self.snapshot.name,
            description=self.snapshot.description,
            metadata=None,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_snapshot_create_with_remote_source(self):
        arglist = [
            '--remote-source',
            'source-name=test_source_name',
            '--remote-source',
            'source-id=test_source_id',
            '--volume',
            self.snapshot.volume_id,
            self.snapshot.name,
        ]
        ref_dict = {
            'source-name': 'test_source_name',
            'source-id': 'test_source_id',
        }
        verifylist = [
            ('remote_source', ref_dict),
            ('volume', self.snapshot.volume_id),
            ("snapshot_name", self.snapshot.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)
        self.volume_client.find_volume.assert_called_once_with(
            self.snapshot.volume_id, ignore_missing=False
        )
        self.volume_client.manage_snapshot.assert_called_with(
            volume_id=self.snapshot.volume_id,
            ref=ref_dict,
            name=self.snapshot.name,
            description=None,
            metadata=None,
        )
        self.volume_client.create_snapshot.assert_not_called()

    def test_snapshot_create_pre_v366(self):
        self.set_volume_api_version('3.65')

        arglist = ["--force", self.snapshot.name]
        verifylist = [("force", True), ("snapshot_name", self.snapshot.name)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        # force parameter should be passed
        self.volume_client.create_snapshot.assert_called_with(
            volume_id=self.snapshot.volume_id,
            force=True,
            name=self.snapshot.name,
            description=None,
            metadata=None,
        )

    def test_snapshot_create_v366_or_later(self):
        self.set_volume_api_version('3.66')

        arglist = [self.snapshot.name]
        verifylist = [("force", False), ("snapshot_name", self.snapshot.name)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        # force parameter should not be passed, for >=3.66
        self.volume_client.create_snapshot.assert_called_with(
            volume_id=self.snapshot.volume_id,
            name=self.snapshot.name,
            description=None,
            metadata=None,
        )

    def test_snapshot_create_v366_or_later_with_force(self):
        """--force should be ignored for microversion >= 3.66."""
        self.set_volume_api_version('3.66')

        arglist = ["--force", self.snapshot.name]
        verifylist = [("force", True), ("snapshot_name", self.snapshot.name)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        # passed but ignored
        self.volume_client.create_snapshot.assert_called_with(
            volume_id=self.snapshot.volume_id,
            name=self.snapshot.name,
            description=None,
            metadata=None,
        )


class TestVolumeSnapshotDelete(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.snapshots = list(
            sdk_fakes.generate_fake_resources(_snapshot.Snapshot)
        )
        self.volume_client.find_snapshot.side_effect = self.snapshots
        self.volume_client.delete_snapshot.return_value = None
        self.volume_client.unmanage_snapshot.return_value = None

        self.cmd = volume_snapshot.DeleteVolumeSnapshot(self.app, None)

    def test_snapshot_delete(self):
        arglist = [self.snapshots[0].id]
        verifylist = [("snapshots", [self.snapshots[0].id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_client.find_snapshot.assert_called_once_with(
            self.snapshots[0].id, ignore_missing=False
        )
        self.volume_client.delete_snapshot.assert_called_once_with(
            self.snapshots[0].id, force=False
        )

    def test_snapshot_delete_with_force(self):
        arglist = ['--force', self.snapshots[0].id]
        verifylist = [('force', True), ("snapshots", [self.snapshots[0].id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_client.find_snapshot.assert_called_once_with(
            self.snapshots[0].id, ignore_missing=False
        )
        self.volume_client.delete_snapshot.assert_called_once_with(
            self.snapshots[0].id, force=True
        )

    def test_delete_multiple_snapshots(self):
        arglist = []
        for s in self.snapshots:
            arglist.append(s.id)
        verifylist = [
            ('snapshots', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_client.find_snapshot.assert_has_calls(
            [mock.call(x.id, ignore_missing=False) for x in self.snapshots]
        )
        self.volume_client.delete_snapshot.assert_has_calls(
            [mock.call(x.id, force=False) for x in self.snapshots]
        )

    def test_delete_multiple_snapshots_with_exception(self):
        self.volume_client.find_snapshot.side_effect = [
            self.snapshots[0],
            sdk_exceptions.NotFoundException(),
        ]

        arglist = [
            self.snapshots[0].id,
            'unexist_snapshot',
        ]
        verifylist = [
            ('snapshots', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )
        self.assertEqual('1 of 2 snapshots failed to delete.', str(exc))

        self.volume_client.find_snapshot.assert_has_calls(
            [
                mock.call(self.snapshots[0].id, ignore_missing=False),
                mock.call('unexist_snapshot', ignore_missing=False),
            ]
        )
        self.volume_client.delete_snapshot.assert_has_calls(
            [
                mock.call(self.snapshots[0].id, force=False),
            ]
        )

    def test_snapshot_delete_remote(self):
        arglist = ['--remote', self.snapshots[0].id]
        verifylist = [('remote', True), ("snapshots", [self.snapshots[0].id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_client.unmanage_snapshot.assert_called_with(
            self.snapshots[0].id
        )

    def test_snapshot_delete_with_remote_force(self):
        arglist = ['--remote', '--force', self.snapshots[0].id]
        verifylist = [
            ('remote', True),
            ('force', True),
            ("snapshots", [self.snapshots[0].id]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            "The --force option is not supported with the --remote parameter.",
            str(exc),
        )

    def test_delete_multiple_snapshots_remote(self):
        arglist = ['--remote']
        for s in self.snapshots:
            arglist.append(s.id)
        verifylist = [('remote', True), ('snapshots', arglist[1:])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_client.unmanage_snapshot.assert_has_calls(
            [mock.call(s.id) for s in self.snapshots]
        )


class TestVolumeSnapshotList(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.volume = sdk_fakes.generate_fake_resource(_volume.Volume)
        self.snapshots = list(
            sdk_fakes.generate_fake_resources(
                _snapshot.Snapshot, attrs={'volume_id': self.volume.name}
            )
        )
        self.volume_client.volumes.return_value = [self.volume]
        self.volume_client.find_volume.return_value = self.volume
        self.volume_client.snapshots.return_value = self.snapshots

        self.columns = ("ID", "Name", "Description", "Status", "Size")
        self.columns_long = (
            *self.columns,
            "Created At",
            "Volume",
            "Properties",
        )

        self.data = []
        self.data_long = []
        for s in self.snapshots:
            self.data.append(
                (
                    s.id,
                    s.name,
                    s.description,
                    s.status,
                    s.size,
                )
            )
            self.data_long.append(
                (
                    s.id,
                    s.name,
                    s.description,
                    s.status,
                    s.size,
                    s.created_at,
                    volume_snapshot.VolumeIdColumn(
                        s.volume_id, volume_cache={self.volume.id: self.volume}
                    ),
                    format_columns.DictColumn(s.metadata),
                )
            )

        self.cmd = volume_snapshot.ListVolumeSnapshot(self.app, None)

    def test_snapshot_list_without_options(self):
        arglist = []
        verifylist = [('all_projects', False), ('long', False)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.snapshots.assert_called_once_with(
            limit=None,
            marker=None,
            max_items=None,
            all_projects=False,
            name=None,
            status=None,
            project_id=None,
            volume_id=None,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_snapshot_list_with_options(self):
        project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = project

        arglist = [
            "--long",
            "--limit",
            "2",
            "--project",
            project.id,
            "--marker",
            self.snapshots[0].id,
        ]
        verifylist = [
            ("long", True),
            ("limit", 2),
            ("project", project.id),
            ("marker", self.snapshots[0].id),
            ('all_projects', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.snapshots.assert_called_once_with(
            limit=2,
            marker=self.snapshots[0].id,
            max_items=None,
            all_projects=True,
            project_id=project.id,
            name=None,
            status=None,
            volume_id=None,
        )
        self.assertEqual(self.columns_long, columns)
        self.assertEqual(self.data_long, list(data))

    def test_snapshot_list_all_projects(self):
        arglist = [
            '--all-projects',
        ]
        verifylist = [('long', False), ('all_projects', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.snapshots.assert_called_once_with(
            limit=None,
            marker=None,
            max_items=None,
            all_projects=True,
            name=None,
            status=None,
            project_id=None,
            volume_id=None,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_snapshot_list_name_option(self):
        arglist = [
            '--name',
            self.snapshots[0].name,
        ]
        verifylist = [
            ('all_projects', False),
            ('long', False),
            ('name', self.snapshots[0].name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.snapshots.assert_called_once_with(
            limit=None,
            marker=None,
            max_items=None,
            all_projects=False,
            name=self.snapshots[0].name,
            status=None,
            project_id=None,
            volume_id=None,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_snapshot_list_status_option(self):
        arglist = [
            '--status',
            'available',
        ]
        verifylist = [
            ('all_projects', False),
            ('long', False),
            ('status', 'available'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.snapshots.assert_called_once_with(
            limit=None,
            marker=None,
            max_items=None,
            all_projects=False,
            name=None,
            status='available',
            project_id=None,
            volume_id=None,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_snapshot_list_volumeid_option(self):
        arglist = [
            '--volume',
            self.volume.id,
        ]
        verifylist = [
            ('all_projects', False),
            ('long', False),
            ('volume', self.volume.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.snapshots.assert_called_once_with(
            limit=None,
            marker=None,
            max_items=None,
            all_projects=False,
            name=None,
            status=None,
            project_id=None,
            volume_id=self.volume.id,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_snapshot_list_negative_limit(self):
        arglist = [
            "--limit",
            "-2",
        ]
        verifylist = [
            ("limit", -2),
        ]
        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )


class TestVolumeSnapshotSet(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.snapshot = sdk_fakes.generate_fake_resource(
            _snapshot.Snapshot, metadata={'foo': 'bar'}
        )
        self.volume_client.find_snapshot.return_value = self.snapshot
        self.volume_client.set_snapshot_metadata.return_value = None
        self.volume_client.update_snapshot.return_value = None
        # Get the command object to mock
        self.cmd = volume_snapshot.SetVolumeSnapshot(self.app, None)

    def test_snapshot_set_no_option(self):
        arglist = [
            self.snapshot.id,
        ]
        verifylist = [
            ("snapshot", self.snapshot.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.volume_client.find_snapshot.assert_called_once_with(
            parsed_args.snapshot, ignore_missing=False
        )
        self.volume_client.reset_snapshot_status.assert_not_called()
        self.volume_client.update_snapshot.assert_not_called()
        self.volume_client.set_snapshot_metadata.assert_not_called()

    def test_snapshot_set_name_and_property(self):
        arglist = [
            "--name",
            "new_snapshot",
            "--property",
            "x=y",
            "--property",
            "foo=foo",
            self.snapshot.id,
        ]
        verifylist = [
            ("name", "new_snapshot"),
            ("properties", {"x": "y", "foo": "foo"}),
            ("snapshot", self.snapshot.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.volume_client.update_snapshot.assert_called_with(
            self.snapshot.id, name="new_snapshot"
        )
        self.volume_client.set_snapshot_metadata.assert_called_with(
            self.snapshot.id, x="y", foo="foo"
        )

    def test_snapshot_set_with_no_property(self):
        arglist = [
            "--no-property",
            self.snapshot.id,
        ]
        verifylist = [
            ("no_property", True),
            ("snapshot", self.snapshot.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.volume_client.find_snapshot.assert_called_once_with(
            parsed_args.snapshot, ignore_missing=False
        )
        self.volume_client.reset_snapshot_status.assert_not_called()
        self.volume_client.update_snapshot.assert_not_called()
        self.volume_client.set_snapshot_metadata.assert_not_called()
        self.volume_client.delete_snapshot_metadata.assert_called_with(
            self.snapshot.id, keys=["foo"]
        )

    def test_snapshot_set_with_no_property_and_property(self):
        arglist = [
            "--no-property",
            "--property",
            "foo_1=bar_1",
            self.snapshot.id,
        ]
        verifylist = [
            ("no_property", True),
            ("properties", {"foo_1": "bar_1"}),
            ("snapshot", self.snapshot.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.volume_client.find_snapshot.assert_called_once_with(
            parsed_args.snapshot, ignore_missing=False
        )
        self.volume_client.reset_snapshot_status.assert_not_called()
        self.volume_client.update_snapshot.assert_not_called()
        self.volume_client.delete_snapshot_metadata.assert_called_with(
            self.snapshot.id, keys=["foo"]
        )
        self.volume_client.set_snapshot_metadata.assert_called_once_with(
            self.snapshot.id,
            foo_1="bar_1",
        )

    def test_snapshot_set_state_to_error(self):
        arglist = ["--state", "error", self.snapshot.id]
        verifylist = [("state", "error"), ("snapshot", self.snapshot.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.volume_client.reset_snapshot_status.assert_called_with(
            self.snapshot.id, "error"
        )

    def test_volume_set_state_failed(self):
        self.volume_client.reset_snapshot_status.side_effect = (
            exceptions.CommandError()
        )
        arglist = ['--state', 'error', self.snapshot.id]
        verifylist = [('state', 'error'), ('snapshot', self.snapshot.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )

        self.assertEqual('One or more of the set operations failed', str(exc))
        self.volume_client.reset_snapshot_status.assert_called_once_with(
            self.snapshot.id, 'error'
        )

    def test_volume_set_name_and_state_failed(self):
        self.volume_client.reset_snapshot_status.side_effect = (
            exceptions.CommandError()
        )
        arglist = [
            '--state',
            'error',
            "--name",
            "new_snapshot",
            self.snapshot.id,
        ]
        verifylist = [
            ('state', 'error'),
            ("name", "new_snapshot"),
            ('snapshot', self.snapshot.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )

        self.assertEqual('One or more of the set operations failed', str(exc))
        self.volume_client.update_snapshot.assert_called_once_with(
            self.snapshot.id, name="new_snapshot"
        )
        self.volume_client.reset_snapshot_status.assert_called_once_with(
            self.snapshot.id, 'error'
        )


class TestVolumeSnapshotShow(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.snapshot = sdk_fakes.generate_fake_resource(_snapshot.Snapshot)

        self.columns = (
            'created_at',
            'description',
            'id',
            'name',
            'properties',
            'size',
            'status',
            'volume_id',
        )
        self.data = (
            self.snapshot.created_at,
            self.snapshot.description,
            self.snapshot.id,
            self.snapshot.name,
            format_columns.DictColumn(self.snapshot.metadata),
            self.snapshot.size,
            self.snapshot.status,
            self.snapshot.volume_id,
        )

        self.volume_client.find_snapshot.return_value = self.snapshot

        self.cmd = volume_snapshot.ShowVolumeSnapshot(self.app, None)

    def test_snapshot_show(self):
        arglist = [self.snapshot.id]
        verifylist = [("snapshot", self.snapshot.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.find_snapshot.assert_called_with(
            self.snapshot.id, ignore_missing=False
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestVolumeSnapshotUnset(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.snapshot = sdk_fakes.generate_fake_resource(_snapshot.Snapshot)
        self.volume_client.find_snapshot.return_value = self.snapshot
        self.volume_client.delete_snapshot_metadata.return_value = None

        self.cmd = volume_snapshot.UnsetVolumeSnapshot(self.app, None)

    def test_snapshot_unset(self):
        arglist = [
            "--property",
            "foo",
            self.snapshot.id,
        ]
        verifylist = [
            ("properties", ["foo"]),
            ("snapshot", self.snapshot.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.volume_client.delete_snapshot_metadata.assert_called_with(
            self.snapshot.id, keys=["foo"]
        )



# Copyright 2019 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from collections.abc import Sequence
import logging
from typing import Any

from manilaclient.common import constants
from osc_lib import exceptions

from openstackclient.common import envvars
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)


def extract_key_value_options(
    pairs: dict[str, str] | None,
) -> dict[str, str]:
    result_dict = {}
    duplicate_options = []
    pairs = pairs or {}

    for attr, value in pairs.items():
        if attr not in result_dict:
            result_dict[attr] = value
        else:
            duplicate_options.append(attr)

    if pairs and len(duplicate_options) > 0:
        duplicate_str = ', '.join(duplicate_options)
        msg = f"Following options were duplicated: {duplicate_str}"
        raise exceptions.CommandError(msg)

    return result_dict


def format_properties(properties: dict[str, str]) -> str:
    formatted_data = []

    for item in properties:
        formatted_data.append(f"{item} : {properties[item]}")
    return "\n".join(formatted_data)


def extract_properties(properties: list[str]) -> dict[str, str]:
    result_dict = {}
    for item in properties:
        try:
            (key, value) = item.split('=', 1)
            if key in result_dict:
                raise exceptions.CommandError(
                    f"Argument '{key}' is specified twice."
                )
            else:
                result_dict[key] = value
        except ValueError:
            raise exceptions.CommandError(
                "Parsing error, expected format 'key=value' for " + item
            )
    return result_dict


def extract_extra_specs(
    extra_specs: dict[str, Any],
    specs_to_add: list[str],
    bool_specs: Sequence[str] = constants.BOOL_SPECS,
) -> dict[str, Any]:
    try:
        for item in specs_to_add:
            (key, value) = item.split('=', 1)
            if key in extra_specs:
                msg = f"Argument '{key}' value specified twice."
                raise exceptions.CommandError(msg)
            elif key in bool_specs:
                if envvars.bool_from_str(value):
                    extra_specs[key] = value.capitalize()
                else:
                    msg = (
                        f"Argument '{key}' is of boolean "
                        f"type and has invalid value: {value!s}"
                    )
                    raise exceptions.CommandError(msg)
            else:
                extra_specs[key] = value
    except ValueError:
        msg = _("Wrong format: specs should be key=value pairs.")
        raise exceptions.CommandError(msg)
    return extra_specs


def extract_group_specs(
    extra_specs: dict[str, Any], specs_to_add: list[str]
) -> dict[str, Any]:
    return extract_extra_specs(
        extra_specs, specs_to_add, constants.GROUP_BOOL_SPECS
    )


def format_column_headers(columns: list[str]) -> list[str]:
    column_headers = []
    for column in columns:
        column_headers.append(
            column.replace('_', ' ').title().replace('Id', 'ID')
        )
    return column_headers


def format_share_group_type(
    share_group_type: Any, formatter: str = 'table'
) -> dict[str, Any]:
    printable_share_group_type = share_group_type._info

    is_public = printable_share_group_type.pop('is_public')

    printable_share_group_type['visibility'] = (
        'public' if is_public else 'private'
    )

    if formatter == 'table':
        printable_share_group_type['group_specs'] = format_properties(
            share_group_type.group_specs
        )
        printable_share_group_type['share_types'] = "\n".join(
            printable_share_group_type['share_types']
        )

    return printable_share_group_type



# Copyright 2020, Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Compute v2 Server action implementations"""

import argparse
from collections.abc import Iterable, Sequence
from typing import Any

from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


def _get_server_share_columns(
    item: Any,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    # Non admin cannot see uuid and export location, so hide them
    if item.uuid is None:
        column_map = {
            'share_id': 'Share ID',
            'status': 'Status',
            'tag': 'Tag',
        }
        hidden_columns = [
            'id',
            'location',
            'name',
            'uuid',
            'export_location',
            'share_proto',
        ]
    else:
        column_map = {
            'uuid': 'UUID',
            'share_id': 'Share ID',
            'status': 'Status',
            'tag': 'Tag',
            'export_location': 'Export Location',
        }
        hidden_columns = ['id', 'location', 'name', 'share_proto']

    return utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden_columns
    )


class ListServerShare(command.Lister):
    """List all the shares attached to a server.

    Requires ``--os-compute-api-version 2.97`` or later.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server',
            metavar='<server>',
            help=_('Server to list share mapping for (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        compute_client = self.app.client_manager.compute

        if not sdk_utils.supports_microversion(compute_client, '2.97'):
            msg = _(
                '--os-compute-api-version 2.97 or greater is required '
                'to support share attachments'
            )
            raise exceptions.CommandError(msg)

        server = compute_client.find_server(
            parsed_args.server,
            ignore_missing=False,
        )
        shares = compute_client.share_attachments(server)

        columns = (
            'share_id',
            'status',
            'tag',
        )
        column_headers = (
            'Share ID',
            'Status',
            'Tag',
        )

        return (
            column_headers,
            (utils.get_item_properties(s, columns) for s in shares),
        )


class ShowServerShare(command.ShowOne):
    """Show detail of a share attachment to a server.

    Requires ``--os-compute-api-version 2.97`` or later.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server',
            metavar='<server>',
            help=_('Server to show share mapping for (name or ID)'),
        )
        parser.add_argument(
            'share',
            metavar='<share>',
            help=_('Share to show details for (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        compute_client = self.app.client_manager.compute
        shared_file_system_client = (
            self.app.client_manager.sdk_connection.shared_file_system
        )

        if not sdk_utils.supports_microversion(compute_client, '2.97'):
            msg = _(
                '--os-compute-api-version 2.97 or greater is required '
                'to support share attachments'
            )
            raise exceptions.CommandError(msg)

        server = compute_client.find_server(
            parsed_args.server,
            ignore_missing=False,
        )
        share = shared_file_system_client.find_share(
            parsed_args.share,
            ignore_missing=False,
        )
        share_attachment = compute_client.get_share_attachment(
            server, share.id
        )

        display_columns, columns = _get_server_share_columns(
            share_attachment,
        )
        data = utils.get_item_properties(share_attachment, columns)
        return display_columns, data


class AddServerShare(command.ShowOne):
    """Add a share to a server.

    Requires ``--os-compute-api-version 2.97`` or later.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server',
            metavar='<server>',
            help=_('Server to add share to (name or ID)'),
        )
        parser.add_argument(
            'share',
            metavar='<share>',
            help=_('Share to add (name or ID)'),
        )
        parser.add_argument(
            '--tag',
            metavar='<tag>',
            help=_(
                'Optional tag used to mount the share, '
                'if not provided the share uuid is used as tag by default'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        compute_client = self.app.client_manager.compute
        shared_file_system_client = (
            self.app.client_manager.sdk_connection.shared_file_system
        )

        if not sdk_utils.supports_microversion(compute_client, '2.97'):
            msg = _(
                '--os-compute-api-version 2.97 or greater is required '
                'to support share attachments'
            )
            raise exceptions.CommandError(msg)

        server = compute_client.find_server(
            parsed_args.server,
            ignore_missing=False,
        )
        share = shared_file_system_client.find_share(
            parsed_args.share,
            ignore_missing=False,
        )

        kwargs: dict[str, Any] = {}
        if parsed_args.tag:
            kwargs['tag'] = parsed_args.tag

        share_attachment = compute_client.create_share_attachment(
            server, share.id, **kwargs
        )

        display_columns, columns = _get_server_share_columns(
            share_attachment,
        )
        data = utils.get_item_properties(share_attachment, columns)
        return display_columns, data


class RemoveServerShare(command.Command):
    """Remove a share from a server.

    Requires ``--os-compute-api-version 2.97`` or later.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server',
            metavar='<server>',
            help=_('Server to remove share from (name or ID)'),
        )
        parser.add_argument(
            'share',
            metavar='<share>',
            help=_('Share to remove (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        compute_client = self.app.client_manager.compute
        shared_file_system_client = (
            self.app.client_manager.sdk_connection.shared_file_system
        )

        if not sdk_utils.supports_microversion(compute_client, '2.97'):
            msg = _(
                '--os-compute-api-version 2.97 or greater is required '
                'to support share attachments'
            )
            raise exceptions.CommandError(msg)

        server = compute_client.find_server(
            parsed_args.server,
            ignore_missing=False,
        )
        share = shared_file_system_client.find_share(
            parsed_args.share,
            ignore_missing=False,
        )
        compute_client.delete_share_attachment(server, share.id)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from manilaclient import api_versions
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils as osc_utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common

LOG = logging.getLogger(__name__)


class CreateShareGroup(command.ShowOne):
    """Create new share group."""

    _description = _("Create new share group")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--name',
            metavar="<name>",
            default=None,
            help=_('Share group name'),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            default=None,
            help=_("Share group description."),
        )
        parser.add_argument(
            "--share-types",
            metavar="<share-types>",
            nargs="+",
            default=[],
            help=_("Name or ID of share type(s)."),
        )
        parser.add_argument(
            "--share-group-type",
            metavar="<share-group-type>",
            default=None,
            help=_(
                "Share group type name or ID of the share group to be created."
            ),
        )
        parser.add_argument(
            "--share-network",
            metavar="<share-network>",
            default=False,
            help=_("Specify share network name or id"),
        )
        parser.add_argument(
            "--source-share-group-snapshot",
            metavar="<source-share-group-snapshot>",
            default=False,
            help=_(
                "Share group snapshot name or ID to create "
                "the share group from."
            ),
        )
        parser.add_argument(
            "--availability-zone",
            metavar='<availability-zone>',
            default=None,
            help=_(
                "Optional availability zone in which group should be created"
            ),
        )
        parser.add_argument(
            "--wait",
            action='store_true',
            default=False,
            help=_('Wait for share group creation'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], tuple[Any, ...]]:
        share_client = self.app.client_manager.share

        share_types = []
        for share_type in parsed_args.share_types:
            share_types.append(
                osc_utils.find_resource(
                    share_client.share_types,
                    share_type,
                )
            )
        share_group_type = None
        if parsed_args.share_group_type:
            share_group_type = osc_utils.find_resource(
                share_client.share_group_types, parsed_args.share_group_type
            ).id

        share_network = None
        if parsed_args.share_network:
            share_network = osc_utils.find_resource(
                share_client.share_networks, parsed_args.share_network
            ).id

        source_share_group_snapshot = None
        if parsed_args.source_share_group_snapshot:
            source_share_group_snapshot = osc_utils.find_resource(
                share_client.share_group_snapshots,
                parsed_args.source_share_group_snapshot,
            ).id

        body = {
            'name': parsed_args.name,
            'description': parsed_args.description,
            'share_types': share_types,
            'share_group_type': share_group_type,
            'share_network': share_network,
            'source_share_group_snapshot': source_share_group_snapshot,
            'availability_zone': parsed_args.availability_zone,
        }

        share_group = share_client.share_groups.create(**body)

        if parsed_args.wait:
            if not osc_utils.wait_for_status(
                status_f=share_client.share_groups.get,
                res_id=share_group.id,
                success_status=['available'],
            ):
                LOG.error(_("ERROR: Share group is in error state."))

            share_group = osc_utils.find_resource(
                share_client.share_groups, share_group.id
            )

        printable_share_group = share_group._info
        printable_share_group.pop('links', None)

        if printable_share_group.get('share_types'):
            if parsed_args.formatter == 'table':
                printable_share_group['share_types'] = "\n".join(
                    printable_share_group['share_types']
                )

        return self.dict2columns(printable_share_group)


class DeleteShareGroup(command.Command):
    """Delete one or more share groups."""

    _description = _("Delete one or more share groups")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "share_group",
            metavar="<share_group>",
            nargs="+",
            help=_("Name or ID of the share group(s) to delete"),
        )
        parser.add_argument(
            "--force",
            action='store_true',
            default=False,
            help=_(
                "Attempt to force delete the share group (Default=False) "
                "(Admin only)."
            ),
        )
        parser.add_argument(
            "--wait",
            action='store_true',
            default=False,
            help=_("Wait for share group to delete"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        share_client = self.app.client_manager.share
        result = 0

        for share_group in parsed_args.share_group:
            try:
                share_group_obj = osc_utils.find_resource(
                    share_client.share_groups, share_group
                )

                share_client.share_groups.delete(
                    share_group_obj, force=parsed_args.force
                )

                if parsed_args.wait:
                    if not osc_utils.wait_for_delete(
                        manager=share_client.share_groups,
                        res_id=share_group_obj.id,
                    ):
                        result += 1

            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete share group with "
                        "name or ID '%(share_group)s': %(e)s"
                    ),
                    {'share_group': share_group, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.share_group)
            msg = _(
                "%(result)s of %(total)s share groups failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListShareGroup(command.Lister):
    """List share groups."""

    _description = _("List share groups")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--all-projects",
            action='store_true',
            default=False,
            help=_("Display share groups from all projects (Admin only)."),
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            default=None,
            help=_("Filter results by name."),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            default=None,
            help=_(
                "Filter results by description. Available "
                "only for microversion >= 2.36."
            ),
        )
        parser.add_argument(
            "--status",
            metavar="<status>",
            default=None,
            help=_("Filter results by status."),
        )
        parser.add_argument(
            "--share-server",
            metavar="<share-server-id>",
            default=None,
            help=_("Filter results by share server ID."),
        )
        parser.add_argument(
            "--share-group-type",
            metavar="<share-group-type>",
            default=None,
            help=_(
                "Filter results by a share group type ID "
                "or name that was used for share group "
                "creation. "
            ),
        )
        parser.add_argument(
            "--snapshot",
            metavar="<snapshot>",
            default=None,
            help=_(
                "Filter results by share group snapshot "
                "name or ID that was used to create the "
                "share group. "
            ),
        )
        parser.add_argument(
            "--host",
            metavar="<host>",
            default=None,
            help=_("Filter results by host."),
        )
        parser.add_argument(
            "--share-network",
            metavar="<share-network>",
            default=None,
            help=_("Filter results by share-network name or ID. "),
        )
        parser.add_argument(
            "--project",
            metavar="<project>",
            default=None,
            help=_(
                "Filter results by project name or ID. Useful with "
                "set key '--all-projects'. "
            ),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        parser.add_argument(
            "--limit",
            metavar="<limit>",
            type=int,
            default=None,
            action=parseractions.NonNegativeAction,
            help=_("Limit the number of share groups returned"),
        )
        parser.add_argument(
            "--marker",
            metavar="<marker>",
            help=_("The last share group ID of the previous page"),
        )
        parser.add_argument(
            '--sort',
            metavar="<key>[:<direction>]",
            default='name:asc',
            help=_(
                "Sort output by selected keys and directions(asc or desc) "
                "(default: name:asc), multiple keys and directions can be "
                "specified separated by comma"
            ),
        )
        parser.add_argument(
            "--name~",
            metavar="<name~>",
            default=None,
            help=_(
                "Filter results matching a share group "
                "name pattern. Available only for "
                "microversion >= 2.36. "
            ),
        )
        parser.add_argument(
            "--description~",
            metavar="<description~>",
            default=None,
            help=_(
                "Filter results matching a share group "
                "description pattern. Available only for "
                "microversion >= 2.36. "
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        share_client = self.app.client_manager.share
        identity_client = self.app.client_manager.identity

        share_server_id = None
        if parsed_args.share_server:
            share_server_id = osc_utils.find_resource(
                share_client.share_servers, parsed_args.share_server
            ).id

        share_group_type = None
        if parsed_args.share_group_type:
            share_group_type = osc_utils.find_resource(
                share_client.share_group_types, parsed_args.share_group_type
            ).id

        snapshot = None
        if parsed_args.snapshot:
            snapshot = osc_utils.find_resource(
                share_client.share_snapshots, parsed_args.snapshot
            ).id

        share_network = None
        if parsed_args.share_network:
            share_network = osc_utils.find_resource(
                share_client.share_networks, parsed_args.share_network
            ).id

        project_id = None
        if parsed_args.project:
            project_id = identity_common.find_project(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            ).id

        columns = [
            'ID',
            'Name',
            'Status',
            'Description',
        ]

        search_opts = {
            'all_tenants': parsed_args.all_projects,
            'name': parsed_args.name,
            'status': parsed_args.status,
            'share_server_id': share_server_id,
            'share_group_type': share_group_type,
            'snapshot': snapshot,
            'host': parsed_args.host,
            'share_network': share_network,
            'project_id': project_id,
            'limit': parsed_args.limit,
            'offset': parsed_args.marker,
        }

        if share_client.api_version >= api_versions.APIVersion("2.36"):
            search_opts['name~'] = getattr(parsed_args, 'name~')
            search_opts['description~'] = getattr(parsed_args, 'description~')
            search_opts['description'] = parsed_args.description
        elif (
            parsed_args.description
            or getattr(parsed_args, 'name~')
            or getattr(parsed_args, 'description~')
        ):
            raise exceptions.CommandError(
                "Pattern based filtering (name~, description~ and description)"
                " is only available with manila API version >= 2.36"
            )

        if parsed_args.all_projects:
            columns.append('Project ID')
        share_groups = share_client.share_groups.list(search_opts=search_opts)

        data = (
            osc_utils.get_dict_properties(share_group._info, columns)
            for share_group in share_groups
        )

        return (columns, data)


class ShowShareGroup(command.ShowOne):
    """Show share group."""

    _description = _("Show details about a share group")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "share_group",
            metavar="<share-group>",
            help=_("Name or ID of the share group."),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], tuple[Any, ...]]:
        share_client = self.app.client_manager.share

        share_group = osc_utils.find_resource(
            share_client.share_groups, parsed_args.share_group
        )

        printable_share_group = share_group._info
        printable_share_group.pop('links', None)

        if printable_share_group.get('share_types'):
            if parsed_args.formatter == 'table':
                printable_share_group['share_types'] = "\n".join(
                    printable_share_group['share_types']
                )

        return self.dict2columns(printable_share_group)


class SetShareGroup(command.Command):
    """Set share group."""

    _description = _("Explicitly set share group status")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share_group',
            metavar="<share-group>",
            help=_('Name or ID of the share group to update.'),
        )
        parser.add_argument(
            '--name',
            metavar="<name>",
            default=None,
            help=_('New name for the share group. (Default=None)'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            default=None,
            help=_('Share group description. (Default=None)'),
        )
        parser.add_argument(
            '--status',
            metavar='<status>',
            default=None,
            help=_(
                'Explicitly update the status of a share group (Admin  '
                'only). Examples include: available, error, creating, '
                'deleting, error_deleting.'
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        share_client = self.app.client_manager.share
        result = 0

        share_group = osc_utils.find_resource(
            share_client.share_groups, parsed_args.share_group
        )

        kwargs = {}
        if parsed_args.name is not None:
            kwargs['name'] = parsed_args.name
        if parsed_args.description is not None:
            kwargs['description'] = parsed_args.description
        if kwargs:
            try:
                share_client.share_groups.update(share_group.id, **kwargs)
            except Exception as e:
                LOG.error(
                    _("Failed to update share group name or description: %s"),
                    e,
                )
                result += 1
        if parsed_args.status:
            try:
                share_group.reset_state(parsed_args.status)
            except Exception as e:
                LOG.error(_("Failed to set status for the share group: %s"), e)
                result += 1

        if result > 0:
            raise exceptions.CommandError(
                _("One or more of the set operations failed")
            )


class UnsetShareGroup(command.Command):
    """Unset a share group property."""

    _description = _("Unset a share group property")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share_group',
            metavar="<share-group>",
            help=_("Name or ID of the share group to set a property for."),
        )
        parser.add_argument(
            "--name",
            action='store_true',
            help=_("Unset share group name."),
        )
        parser.add_argument(
            "--description",
            action='store_true',
            help=_("Unset share group description."),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        share_client = self.app.client_manager.share

        share_group = osc_utils.find_resource(
            share_client.share_groups, parsed_args.share_group
        )

        kwargs: dict[str, None] = {}
        if parsed_args.name:
            kwargs['name'] = None
        if parsed_args.description:
            kwargs['description'] = None
        if kwargs:
            try:
                share_client.share_groups.update(share_group, **kwargs)
            except Exception as e:
                msg = _(
                    "Failed to unset share_group name or description: %(e)s"
                )
                raise exceptions.CommandError(msg % {'e': e})



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

from openstack.block_storage.v3 import (
    manageable_snapshot as _manageable_snapshot,
)
from openstack.block_storage.v3 import manageable_volume as _manageable_volume
from osc_lib import exceptions

from openstackclient.tests.unit import utils as tests_utils
from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes
from openstackclient.volume.v3 import block_storage_manage

FAKE_VOLUME_MANAGE_LIST = [
    _manageable_volume.ManageableVolume(
        reference={'source-name': 'fake-volume'},
        size='1',
        safe_to_manage=False,
        reason_not_safe='already managed',
        cinder_id='fake-volume',
        extra_info=None,
    ),
    _manageable_volume.ManageableVolume(
        reference={'source-name': 'fake-volume'},
        size='2',
        safe_to_manage=False,
        reason_not_safe='already managed',
        cinder_id='fake-volume',
        extra_info=None,
    ),
]

FAKE_SNAPSHOT_MANAGE_LIST = [
    _manageable_snapshot.ManageableSnapshot(
        reference={'source-name': 'fake-snapshot'},
        source_reference={'source-name': 'fake-source'},
        size='1',
        safe_to_manage=False,
        reason_not_safe='already managed',
        cinder_id='fake-snapshot',
        extra_info=None,
    ),
    _manageable_snapshot.ManageableSnapshot(
        reference={'source-name': 'fake-snapshot'},
        source_reference={'source-name': 'fake-source'},
        size='2',
        safe_to_manage=False,
        reason_not_safe='already managed',
        cinder_id='fake-snapshot',
        extra_info=None,
    ),
]


class TestBlockStorageVolumeManage(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.volume_client.manageable_volumes.return_value = iter(
            FAKE_VOLUME_MANAGE_LIST
        )

        self.cmd = block_storage_manage.BlockStorageManageVolumes(
            self.app, None
        )

    def test_block_storage_volume_manage_list(self):
        self.set_volume_api_version('3.8')

        arglist = ['fake_host']
        verifylist = [('host', 'fake_host')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = ['reference', 'size', 'safe_to_manage']
        expected_data = tuple(
            (v.reference, v.size, v.safe_to_manage)
            for v in FAKE_VOLUME_MANAGE_LIST
        )

        self.assertEqual(expected_columns, columns)
        self.assertEqual(expected_data, tuple(data))

        self.volume_client.manageable_volumes.assert_called_once_with(
            details=False, host='fake_host'
        )

    def test_block_storage_volume_manage_list__pre_v38(self):
        arglist = ['fake_host']
        verifylist = [('host', 'fake_host')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.8 or greater is required', str(exc)
        )

    def test_block_storage_volume_manage_list__pre_v317(self):
        self.set_volume_api_version('3.16')

        arglist = ['--cluster', 'fake_cluster']
        verifylist = [('cluster', 'fake_cluster')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.17 or greater is required', str(exc)
        )
        self.assertIn('--cluster', str(exc))

    def test_block_storage_volume_manage_list__host_and_cluster(self):
        self.set_volume_api_version('3.17')

        arglist = ['fake_host', '--cluster', 'fake_cluster']
        verifylist = [('host', 'fake_host'), ('cluster', 'fake_cluster')]
        exc = self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )
        self.assertIn(
            'argument --cluster: not allowed with argument <host>', str(exc)
        )

    def test_block_storage_volume_manage_list__detailed(self):
        """This option is deprecated."""
        self.set_volume_api_version('3.8')

        arglist = ['--detailed', 'True', 'fake_host']
        verifylist = [
            ('host', 'fake_host'),
            ('detailed', 'True'),
            ('marker', None),
            ('limit', None),
            ('offset', None),
            ('sort', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch.object(self.cmd.log, 'warning') as mock_warning:
            columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'reference',
            'size',
            'safe_to_manage',
            'reason_not_safe',
            'cinder_id',
            'extra_info',
        ]
        expected_data = tuple(
            (
                v.reference,
                v.size,
                v.safe_to_manage,
                v.reason_not_safe,
                v.cinder_id,
                v.extra_info,
            )
            for v in FAKE_VOLUME_MANAGE_LIST
        )

        self.assertEqual(expected_columns, columns)
        self.assertEqual(expected_data, tuple(data))

        self.volume_client.manageable_volumes.assert_called_once_with(
            details=True, host='fake_host'
        )
        mock_warning.assert_called_once()
        self.assertIn(
            "The --detailed option has been deprecated.",
            str(mock_warning.call_args[0][0]),
        )

    def test_block_storage_volume_manage_list__all_args(self):
        self.set_volume_api_version('3.8')

        arglist = [
            'fake_host',
            '--long',
            '--marker',
            'fake_marker',
            '--limit',
            '5',
            '--offset',
            '3',
            '--sort',
            'size:asc',
        ]
        verifylist = [
            ('host', 'fake_host'),
            ('detailed', None),
            ('long', True),
            ('marker', 'fake_marker'),
            ('limit', '5'),
            ('offset', '3'),
            ('sort', 'size:asc'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'reference',
            'size',
            'safe_to_manage',
            'reason_not_safe',
            'cinder_id',
            'extra_info',
        ]
        expected_data = tuple(
            (
                v.reference,
                v.size,
                v.safe_to_manage,
                v.reason_not_safe,
                v.cinder_id,
                v.extra_info,
            )
            for v in FAKE_VOLUME_MANAGE_LIST
        )

        self.assertEqual(expected_columns, columns)
        self.assertEqual(expected_data, tuple(data))

        self.volume_client.manageable_volumes.assert_called_once_with(
            details=True,
            host='fake_host',
            marker='fake_marker',
            limit='5',
            offset='3',
            sort='size:asc',
        )


class TestBlockStorageSnapshotManage(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.volume_client.manageable_snapshots.return_value = iter(
            FAKE_SNAPSHOT_MANAGE_LIST
        )

        self.cmd = block_storage_manage.BlockStorageManageSnapshots(
            self.app, None
        )

    def test_block_storage_snapshot_manage_list(self):
        self.set_volume_api_version('3.8')

        arglist = ['fake_host']
        verifylist = [('host', 'fake_host')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'reference',
            'size',
            'safe_to_manage',
            'source_reference',
        ]
        expected_data = tuple(
            (s.reference, s.size, s.safe_to_manage, s.source_reference)
            for s in FAKE_SNAPSHOT_MANAGE_LIST
        )

        self.assertEqual(expected_columns, columns)
        self.assertEqual(expected_data, tuple(data))

        self.volume_client.manageable_snapshots.assert_called_once_with(
            details=False, host='fake_host'
        )

    def test_block_storage_snapshot_manage_list__pre_v38(self):
        arglist = ['fake_host']
        verifylist = [('host', 'fake_host')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.8 or greater is required', str(exc)
        )

    def test_block_storage_snapshot_manage_list__pre_v317(self):
        self.set_volume_api_version('3.16')

        arglist = ['--cluster', 'fake_cluster']
        verifylist = [('cluster', 'fake_cluster')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.17 or greater is required', str(exc)
        )
        self.assertIn('--cluster', str(exc))

    def test_block_storage_snapshot_manage_list__host_and_cluster(self):
        self.set_volume_api_version('3.17')

        arglist = ['fake_host', '--cluster', 'fake_cluster']
        verifylist = [('host', 'fake_host'), ('cluster', 'fake_cluster')]
        exc = self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )
        self.assertIn(
            'argument --cluster: not allowed with argument <host>', str(exc)
        )

    def test_block_storage_snapshot_manage_list__detailed(self):
        self.set_volume_api_version('3.8')

        arglist = ['--detailed', 'True', 'fake_host']
        verifylist = [
            ('host', 'fake_host'),
            ('detailed', 'True'),
            ('marker', None),
            ('limit', None),
            ('offset', None),
            ('sort', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch.object(self.cmd.log, 'warning') as mock_warning:
            columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'reference',
            'size',
            'safe_to_manage',
            'source_reference',
            'reason_not_safe',
            'cinder_id',
            'extra_info',
        ]
        expected_data = tuple(
            (
                s.reference,
                s.size,
                s.safe_to_manage,
                s.source_reference,
                s.reason_not_safe,
                s.cinder_id,
                s.extra_info,
            )
            for s in FAKE_SNAPSHOT_MANAGE_LIST
        )

        self.assertEqual(expected_columns, columns)
        self.assertEqual(expected_data, tuple(data))

        self.volume_client.manageable_snapshots.assert_called_once_with(
            details=True, host='fake_host'
        )
        mock_warning.assert_called_once()
        self.assertIn(
            "The --detailed option has been deprecated.",
            str(mock_warning.call_args[0][0]),
        )

    def test_block_storage_snapshot_manage_list__all_args(self):
        self.set_volume_api_version('3.8')

        arglist = [
            '--long',
            '--marker',
            'fake_marker',
            '--limit',
            '5',
            '--offset',
            '3',
            '--sort',
            'size:asc',
            'fake_host',
        ]
        verifylist = [
            ('host', 'fake_host'),
            ('detailed', None),
            ('long', True),
            ('marker', 'fake_marker'),
            ('limit', '5'),
            ('offset', '3'),
            ('sort', 'size:asc'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'reference',
            'size',
            'safe_to_manage',
            'source_reference',
            'reason_not_safe',
            'cinder_id',
            'extra_info',
        ]
        expected_data = tuple(
            (
                s.reference,
                s.size,
                s.safe_to_manage,
                s.source_reference,
                s.reason_not_safe,
                s.cinder_id,
                s.extra_info,
            )
            for s in FAKE_SNAPSHOT_MANAGE_LIST
        )

        self.assertEqual(expected_columns, columns)
        self.assertEqual(expected_data, tuple(data))

        self.volume_client.manageable_snapshots.assert_called_once_with(
            details=True,
            host='fake_host',
            marker='fake_marker',
            limit='5',
            offset='3',
            sort='size:asc',
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest.mock import call

from osc_lib.cli import format_columns
from osc_lib import exceptions

from openstackclient.network.v2 import network_agent
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as tests_utils


class TestNetworkAgent(network_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()


class TestAddNetworkToAgent(TestNetworkAgent):
    net = network_fakes.create_one_network()
    agent = network_fakes.create_one_network_agent()

    def setUp(self):
        super().setUp()

        self.network_client.get_agent.return_value = self.agent
        self.network_client.find_network.return_value = self.net
        self.network_client.name = self.network_client.find_network.name

        self.cmd = network_agent.AddNetworkToAgent(self.app, None)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_add_network_to_dhcp_agent(self):
        arglist = ['--dhcp', self.agent.id, self.net.id]
        verifylist = [
            ('dhcp', True),
            ('agent_id', self.agent.id),
            ('network', self.net.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.network_client.add_dhcp_agent_to_network.assert_called_once_with(
            self.agent, self.net
        )


class TestAddRouterAgent(TestNetworkAgent):
    _router = network_fakes.create_one_router()
    _agent = network_fakes.create_one_network_agent()

    def setUp(self):
        super().setUp()

        self.network_client.get_agent.return_value = self._agent
        self.network_client.find_router.return_value = self._router

        self.cmd = network_agent.AddRouterToAgent(self.app, None)

    def test_add_no_options(self):
        arglist = []
        verifylist = []

        # Missing agent ID will cause command to bail
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_add_router_required_options(self):
        arglist = [
            self._agent.id,
            self._router.id,
            '--l3',
        ]
        verifylist = [
            ('l3', True),
            ('agent_id', self._agent.id),
            ('router', self._router.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.add_router_to_agent.assert_called_with(
            self._agent, self._router, ha_chassis_priority=None
        )
        self.assertIsNone(result)

    def test_add_router_with_ha_chassis_priority(self):
        arglist = [
            self._agent.id,
            self._router.id,
            '--l3',
            '--ha-chassis-priority',
            '100',
        ]
        verifylist = [
            ('l3', True),
            ('agent_id', self._agent.id),
            ('router', self._router.id),
            ('ha_chassis_priority', 100),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.add_router_to_agent.assert_called_with(
            self._agent,
            self._router,
            ha_chassis_priority=100,
        )
        self.assertIsNone(result)


class TestDeleteNetworkAgent(TestNetworkAgent):
    network_agents = network_fakes.create_network_agents(count=2)

    def setUp(self):
        super().setUp()
        self.network_client.delete_agent.return_value = None

        # Get the command object to test
        self.cmd = network_agent.DeleteNetworkAgent(self.app, None)

    def test_network_agent_delete(self):
        arglist = [
            self.network_agents[0].id,
        ]
        verifylist = [
            ('network_agent', [self.network_agents[0].id]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.network_client.delete_agent.assert_called_once_with(
            self.network_agents[0].id, ignore_missing=False
        )
        self.assertIsNone(result)

    def test_multi_network_agents_delete(self):
        arglist = []

        for n in self.network_agents:
            arglist.append(n.id)
        verifylist = [
            ('network_agent', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        calls = []
        for n in self.network_agents:
            calls.append(call(n.id, ignore_missing=False))
        self.network_client.delete_agent.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_multi_network_agents_delete_with_exception(self):
        arglist = [
            self.network_agents[0].id,
            'unexist_network_agent',
        ]
        verifylist = [
            (
                'network_agent',
                [self.network_agents[0].id, 'unexist_network_agent'],
            ),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        delete_mock_result = [True, exceptions.CommandError]
        self.network_client.delete_agent.side_effect = delete_mock_result

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 network agents failed to delete.', str(e))

        self.network_client.delete_agent.assert_any_call(
            self.network_agents[0].id, ignore_missing=False
        )
        self.network_client.delete_agent.assert_any_call(
            'unexist_network_agent', ignore_missing=False
        )


class TestListNetworkAgent(TestNetworkAgent):
    network_agents = network_fakes.create_network_agents(count=3)

    columns = (
        'ID',
        'Agent Type',
        'Host',
        'Availability Zone',
        'Alive',
        'State',
        'Binary',
    )
    data = []
    for agent in network_agents:
        data.append(
            (
                agent.id,
                agent.agent_type,
                agent.host,
                agent.availability_zone,
                network_agent.AliveColumn(agent.is_alive),
                network_agent.AdminStateColumn(agent.is_admin_state_up),
                agent.binary,
            )
        )

    def setUp(self):
        super().setUp()

        self.network_client.agents.return_value = self.network_agents
        self.network_client.routers_hosting_l3_agents.return_value = (
            self.network_agents
        )

        _testagent = network_fakes.create_one_network_agent()
        self.network_client.get_agent.return_value = _testagent
        self.network_client.get_agent.return_value = _testagent

        self._testnetwork = network_fakes.create_one_network()
        self.network_client.find_network.return_value = self._testnetwork
        self.network_client.network_hosting_dhcp_agents.return_value = (
            self.network_agents
        )

        self._testrouter = network_fakes.create_one_router()
        self.network_client.find_router.return_value = self._testrouter

        # Get the command object to test
        self.cmd = network_agent.ListNetworkAgent(self.app, None)

    def test_network_agents_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.agents.assert_called_once_with(**{})
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_network_agents_list_pagination(self):
        arglist = [
            '--marker',
            self.network_agents[0].id,
            '--limit',
            '1',
        ]
        verifylist = [
            ('marker', self.network_agents[0].id),
            ('limit', 1),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.agents.assert_called_once_with(
            **{
                'marker': self.network_agents[0].id,
                'limit': 1,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_network_agents_list_agent_type(self):
        arglist = [
            '--agent-type',
            'dhcp',
        ]
        verifylist = [
            ('agent_type', 'dhcp'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.agents.assert_called_once_with(
            **{
                'agent_type': 'DHCP agent',
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_network_agents_list_host(self):
        arglist = [
            '--host',
            self.network_agents[0].host,
        ]
        verifylist = [
            ('host', self.network_agents[0].host),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.agents.assert_called_once_with(
            **{
                'host': self.network_agents[0].host,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_network_agents_list_networks(self):
        arglist = [
            '--network',
            self._testnetwork.id,
        ]
        verifylist = [
            ('network', self._testnetwork.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.network_hosting_dhcp_agents.assert_called_once_with(
            self._testnetwork
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_network_agents_list_routers(self):
        arglist = [
            '--router',
            self._testrouter.id,
        ]
        verifylist = [('router', self._testrouter.id), ('long', False)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.routers_hosting_l3_agents.assert_called_once_with(
            self._testrouter
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_network_agents_list_routers_with_long_option(self):
        arglist = [
            '--router',
            self._testrouter.id,
            '--long',
        ]
        verifylist = [('router', self._testrouter.id), ('long', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.routers_hosting_l3_agents.assert_called_once_with(
            self._testrouter
        )

        # Add a column 'HA State' and corresponding data.
        router_agent_columns = (*self.columns, 'HA State')
        router_agent_data = [(*d, '') for d in self.data]

        self.assertEqual(router_agent_columns, columns)
        self.assertEqual(len(router_agent_data), len(list(data)))


class TestRemoveNetworkFromAgent(TestNetworkAgent):
    net = network_fakes.create_one_network()
    agent = network_fakes.create_one_network_agent()

    def setUp(self):
        super().setUp()

        self.network_client.get_agent.return_value = self.agent
        self.network_client.find_network.return_value = self.net
        self.network_client.name = self.network_client.find_network.name

        self.cmd = network_agent.RemoveNetworkFromAgent(self.app, None)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_network_agents_list_routers_no_arg(self):
        arglist = [
            '--routers',
        ]
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_network_from_dhcp_agent(self):
        arglist = ['--dhcp', self.agent.id, self.net.id]
        verifylist = [
            ('dhcp', True),
            ('agent_id', self.agent.id),
            ('network', self.net.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.network_client.remove_dhcp_agent_from_network.assert_called_once_with(
            self.agent, self.net
        )


class TestRemoveRouterAgent(TestNetworkAgent):
    _router = network_fakes.create_one_router()
    _agent = network_fakes.create_one_network_agent()

    def setUp(self):
        super().setUp()

        self.network_client.get_agent.return_value = self._agent
        self.network_client.find_router.return_value = self._router

        self.cmd = network_agent.RemoveRouterFromAgent(self.app, None)

    def test_remove_no_options(self):
        arglist = []
        verifylist = []

        # Missing agent ID will cause command to bail
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_remove_router_required_options(self):
        arglist = [
            '--l3',
            self._agent.id,
            self._router.id,
        ]
        verifylist = [
            ('l3', True),
            ('agent_id', self._agent.id),
            ('router', self._router.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.remove_router_from_agent.assert_called_with(
            self._agent, self._router
        )
        self.assertIsNone(result)


class TestSetNetworkAgentRouter(TestNetworkAgent):
    def setUp(self):
        super().setUp()
        self._router = network_fakes.create_one_router()
        self._agent = network_fakes.create_one_network_agent()
        self.network_client.get_agent.return_value = self._agent
        self.network_client.find_router.return_value = self._router
        self.cmd = network_agent.SetNetworkAgentRouter(self.app, None)

    def test_set_no_options(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_set_router_missing_ha_chassis_priority(self):
        arglist = [
            self._agent.id,
            self._router.id,
        ]
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_set_router_with_ha_chassis_priority(self):
        arglist = [
            '--ha-chassis-priority',
            '200',
            self._agent.id,
            self._router.id,
        ]
        verifylist = [
            ('agent_id', self._agent.id),
            ('router', self._router.id),
            ('ha_chassis_priority', 200),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_router_in_agent.assert_called_with(
            self._agent,
            self._router,
            ha_chassis_priority=200,
        )
        self.assertIsNone(result)


class TestSetNetworkAgent(TestNetworkAgent):
    _network_agent = network_fakes.create_one_network_agent()

    def setUp(self):
        super().setUp()
        self.network_client.update_agent.return_value = None
        self.network_client.get_agent.return_value = self._network_agent

        # Get the command object to test
        self.cmd = network_agent.SetNetworkAgent(self.app, None)

    def test_set_nothing(self):
        arglist = [
            self._network_agent.id,
        ]
        verifylist = [
            ('network_agent', self._network_agent.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {}
        self.network_client.update_agent.assert_called_once_with(
            self._network_agent, **attrs
        )
        self.assertIsNone(result)

    def test_set_all(self):
        arglist = [
            '--description',
            'new_description',
            '--enable',
            self._network_agent.id,
        ]
        verifylist = [
            ('description', 'new_description'),
            ('enable', True),
            ('disable', False),
            ('network_agent', self._network_agent.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {
            'description': 'new_description',
            'admin_state_up': True,
            'is_admin_state_up': True,
        }
        self.network_client.update_agent.assert_called_once_with(
            self._network_agent, **attrs
        )
        self.assertIsNone(result)

    def test_set_with_disable(self):
        arglist = [
            '--disable',
            self._network_agent.id,
        ]
        verifylist = [
            ('enable', False),
            ('disable', True),
            ('network_agent', self._network_agent.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {
            'admin_state_up': False,
            'is_admin_state_up': False,
        }
        self.network_client.update_agent.assert_called_once_with(
            self._network_agent, **attrs
        )
        self.assertIsNone(result)


class TestShowNetworkAgent(TestNetworkAgent):
    _network_agent = network_fakes.create_one_network_agent()

    columns = (
        'admin_state_up',
        'agent_type',
        'alive',
        'availability_zone',
        'binary',
        'configuration',
        'created_at',
        'description',
        'ha_chassis_priority',
        'ha_state',
        'host',
        'id',
        'last_heartbeat_at',
        'resources_synced',
        'started_at',
        'topic',
    )
    data = (
        network_agent.AdminStateColumn(_network_agent.is_admin_state_up),
        _network_agent.agent_type,
        network_agent.AliveColumn(_network_agent.is_alive),
        _network_agent.availability_zone,
        _network_agent.binary,
        format_columns.DictColumn(_network_agent.configuration),
        _network_agent.created_at,
        _network_agent.description,
        # NOTE(ralonsoh): `ha_chassis_priority` column is still not supported
        # by the API. See LP#2103521.
        None,
        _network_agent.ha_state,
        _network_agent.host,
        _network_agent.id,
        _network_agent.last_heartbeat_at,
        _network_agent.resources_synced,
        _network_agent.started_at,
        _network_agent.topic,
    )

    def setUp(self):
        super().setUp()
        self.network_client.get_agent.return_value = self._network_agent

        # Get the command object to test
        self.cmd = network_agent.ShowNetworkAgent(self.app, None)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_show_all_options(self):
        arglist = [
            self._network_agent.id,
        ]
        verifylist = [
            ('network_agent', self._network_agent.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.get_agent.assert_called_once_with(
            self._network_agent.id
        )
        self.assertCountEqual(self.columns, columns)
        self.assertEqual(len(list(self.data)), len(list(data)))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import tempfile

from tempest.lib.common.utils import data_utils
from tempest.lib import exceptions

from openstackclient.tests.functional import base


class KeypairBase(base.TestCase):
    """Methods for functional tests."""

    def keypair_create(self, name=data_utils.rand_uuid(), user=None):
        """Create keypair and add cleanup."""
        cmd = 'keypair create ' + name
        if user is not None:
            cmd += ' --user ' + user
        raw_output = self.openstack(cmd)
        self.addCleanup(
            self.keypair_delete, name, ignore_exceptions=True, user=user
        )
        if not raw_output:
            self.fail('Keypair has not been created!')
        return name

    def keypair_list(self, params=''):
        """Return dictionary with list of keypairs."""
        raw_output = self.openstack('keypair list')
        keypairs = self.parse_show_as_object(raw_output)
        return keypairs

    def keypair_delete(self, name, ignore_exceptions=False, user=None):
        """Try to delete keypair by name."""
        try:
            cmd = 'keypair delete ' + name
            if user is not None:
                cmd += ' --user ' + user
            self.openstack(cmd)
        except exceptions.CommandFailed:
            if not ignore_exceptions:
                raise


class KeypairTests(KeypairBase):
    """Functional tests for compute keypairs."""

    PUBLIC_KEY = (
        'ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDWNGczJxNaFUrJJVhta4dWsZY6bU'
        '5HUMPbyfSMu713ca3mYtG848W4dfDCB98KmSQx2Bl0D6Q2nrOszOXEQWAXNdfMadnW'
        'c4mNwhZcPBVohIFoC1KZJC8kcBTvFZcoz3mdIijxJtywZNpGNh34VRJlZeHyYjg8/D'
        'esHzdoBVd5c/4R36emQSIV9ukY6PHeZ3scAH4B3K9PxItJBwiFtouSRphQG0bJgOv/'
        'gjAjMElAvg5oku98cb4QiHZ8T8WY68id804raHR6pJxpVVJN4TYJmlUs+NOVM+pPKb'
        'KJttqrIBTkawGK9pLHNfn7z6v1syvUo/4enc1l0Q/Qn2kWiz67 fake@openstack'
    )

    def setUp(self):
        """Create keypair with randomized name for tests."""
        super().setUp()
        self.KPName = data_utils.rand_name('TestKeyPair')
        self.keypair = self.keypair_create(self.KPName)

    def test_keypair_create_duplicate(self):
        """Try to create duplicate name keypair.

        Test steps:
        1) Create keypair in setUp
        2) Try to create duplicate keypair with the same name
        """
        self.assertRaises(
            exceptions.CommandFailed,
            self.openstack,
            'keypair create ' + self.KPName,
        )

    def test_keypair_create_noname(self):
        """Try to create keypair without name.

        Test steps:
        1) Try to create keypair without a name
        """
        self.assertRaises(
            exceptions.CommandFailed, self.openstack, 'keypair create'
        )

    def test_keypair_create_public_key(self):
        """Test for create keypair with --public-key option.

        Test steps:
        1) Create keypair with given public key
        2) Delete keypair
        """
        with tempfile.NamedTemporaryFile(mode='w+') as f:
            f.write(self.PUBLIC_KEY)
            f.flush()

            raw_output = self.openstack(
                f'keypair create --public-key {f.name} tmpkey',
            )
            self.addCleanup(
                self.openstack,
                'keypair delete tmpkey',
            )
            self.assertIn('tmpkey', raw_output)

    def test_keypair_create_private_key(self):
        """Test for create keypair with --private-key option.

        Test steps:
        1) Create keypair with private key file
        2) Delete keypair
        """
        with tempfile.NamedTemporaryFile(mode='w+') as f:
            cmd_output = self.openstack(
                f'keypair create --private-key {f.name} tmpkey',
                parse_output=True,
            )
            self.addCleanup(self.openstack, 'keypair delete tmpkey')
            self.assertEqual('tmpkey', cmd_output.get('name'))
            self.assertIsNotNone(cmd_output.get('user_id'))
            self.assertIsNotNone(cmd_output.get('fingerprint'))
            pk_content = f.read()
            self.assertInOutput(
                '-----BEGIN OPENSSH PRIVATE KEY-----',
                pk_content,
            )
            self.assertRegex(pk_content, "[0-9A-Za-z+/]+[=]{0,3}\n")
            self.assertInOutput(
                '-----END OPENSSH PRIVATE KEY-----',
                pk_content,
            )

    def test_keypair_create(self):
        """Test keypair create command.

        Test steps:
        1) Create keypair in setUp
        2) Check Ed25519 private key in output
        3) Check for new keypair in keypairs list
        """
        NewName = data_utils.rand_name('TestKeyPairCreated')
        raw_output = self.openstack('keypair create ' + NewName)
        self.addCleanup(self.openstack, 'keypair delete ' + NewName)
        self.assertInOutput('-----BEGIN OPENSSH PRIVATE KEY-----', raw_output)
        self.assertRegex(raw_output, "[0-9A-Za-z+/]+[=]{0,3}\n")
        self.assertInOutput('-----END OPENSSH PRIVATE KEY-----', raw_output)
        self.assertIn(NewName, self.keypair_list())

    def test_keypair_delete_not_existing(self):
        """Try to delete keypair with not existing name.

        Test steps:
        1) Create keypair in setUp
        2) Try to delete not existing keypair
        """
        self.assertRaises(
            exceptions.CommandFailed,
            self.openstack,
            'keypair delete not_existing',
        )

    def test_keypair_delete(self):
        """Test keypair delete command.

        Test steps:
        1) Create keypair in setUp
        2) Delete keypair
        3) Check that keypair not in keypairs list
        """
        self.openstack('keypair delete ' + self.KPName)
        self.assertNotIn(self.KPName, self.keypair_list())

    def test_keypair_list(self):
        """Test keypair list command.

        Test steps:
        1) Create keypair in setUp
        2) List keypairs
        3) Check output table structure
        4) Check keypair name in output
        """
        HEADERS = ['Name', 'Fingerprint']
        raw_output = self.openstack('keypair list')
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, HEADERS)
        self.assertIn(self.KPName, raw_output)

    def test_keypair_show(self):
        """Test keypair show command.

        Test steps:
        1) Create keypair in setUp
        2) Show keypair
        3) Check output table structure
        4) Check keypair name in output
        """
        HEADERS = ['Field', 'Value']
        raw_output = self.openstack('keypair show ' + self.KPName)
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, HEADERS)
        self.assertInOutput(self.KPName, raw_output)

    def test_keypair_list_by_project(self):
        """Test keypair list by project.

        Test steps:
        1) Create keypair for admin project in setUp
        2) Create a new project
        3) Create a new user
        4) Associate the new user with the new project
        5) Create keypair for the new user
        6) List keypairs by the new project
        7) Check that only the keypair from step 5 is returned
        """
        project_name = data_utils.rand_name('TestProject')
        self.openstack(f'project create {project_name}')
        self.addCleanup(self.openstack, f'project delete {project_name}')
        user_name = data_utils.rand_name('TestUser')
        self.openstack(f'user create {user_name}')
        self.addCleanup(self.openstack, f'user delete {user_name}')
        self.openstack(
            f'role add --user {user_name} --project {project_name} member'
        )
        keypair_name = self.keypair_create(user=user_name)
        raw_output = self.openstack(f'keypair list --project {project_name}')
        items = self.parse_listing(raw_output)
        self.assertEqual(1, len(items))
        self.assertEqual(keypair_name, items[0]['Name'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import tempfile

from openstackclient.tests.functional.identity.v3 import common


class PolicyTests(common.IdentityTests):
    def test_policy_create(self):
        self._create_dummy_policy()

    def test_policy_delete(self):
        policy = self._create_dummy_policy(add_clean_up=False)
        raw_output = self.openstack(f'policy delete {policy}')
        self.assertEqual(0, len(raw_output))

    def test_policy_multi_delete(self):
        policy_1 = self._create_dummy_policy(add_clean_up=False)
        policy_2 = self._create_dummy_policy(add_clean_up=False)
        raw_output = self.openstack(f'policy delete {policy_1} {policy_2}')
        self.assertEqual(0, len(raw_output))

    def test_policy_show(self):
        policy = self._create_dummy_policy(add_clean_up=True)
        raw_output = self.openstack(f'policy show {policy}')
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.POLICY_FIELDS)

    def test_policy_list(self):
        self._create_dummy_policy(add_clean_up=True)
        raw_output = self.openstack('policy list')
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, self.POLICY_LIST_HEADERS)

    def test_policy_list_long(self):
        self._create_dummy_policy(add_clean_up=True)
        raw_output = self.openstack('policy list --long')
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, self.POLICY_LIST_LONG_HEADERS)

    def test_policy_set(self):
        policy = self._create_dummy_policy(add_clean_up=True)
        with tempfile.NamedTemporaryFile(mode='w+') as f:
            NEW_RULES = [
                {
                    "local": [{"group": {"id": "85a868"}}],
                    "remote": [
                        {"type": "orgPersonType", "any_one_of": ["Employee"]},
                        {"type": "sn", "any_one_of": ["Young"]},
                    ],
                },
                {
                    "local": [
                        {"group": {"id": "0cd5e9"}},
                        {"user": {"name": "0cd5e9"}},
                    ],
                    "remote": [
                        {"type": "UserName"},
                        {
                            "type": "orgPersonType",
                            "not_any_of": ["Contractor", "SubContractor"],
                        },
                        {"type": "LastName", "any_one_of": ["Bo"]},
                    ],
                },
            ]
            f.write(json.dumps(NEW_RULES))
            f.flush()
            raw_output = self.openstack(
                f'policy set {policy} --rules {f.name} --type text/json'
            )
        self.assertEqual(0, len(raw_output))
        raw_output = self.openstack(f'policy show {policy}')
        updated_value = self.parse_show_as_object(raw_output)
        self.assertEqual('text/json', updated_value['type'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ast
import os
import re

from hacking import core

"""
Guidelines for writing new hacking checks

 - Use only for python-openstackclient specific tests. OpenStack general tests
   should be submitted to the common 'hacking' module.
 - Pick numbers in the range O4xx. Find the current test with the highest
   allocated number and then pick the next value.
"""


@core.flake8ext
def assert_no_oslo(logical_line):
    """Check for use of oslo libraries.

    Okay: import os
    Okay: from os import path
    O400: import oslo_messaging
    O400: from oslo_log import log
    """
    if match := re.match(r'(from|import) (oslo_.*)', logical_line):
        if match.group(2) == 'oslo_i18n':
            return
        yield (0, "O400: oslo libraries should not be used in SDK projects")


@core.flake8ext
def assert_no_duplicated_setup(logical_line, filename):
    """Check for use of various unnecessary test duplications.

    This check only applies to files under openstackclient/tests/unit/.

    Okay: self.app = fakes.FakeShell()
    Okay: self.app.client_manager.auth_ref = mock.Mock()
    O401: self.app = Namespace(self.app, self.namespace)
    O401: self.network_client = self.app.client_manager.network
    O401: self.app.client_manager.network = mock.Mock()
    """
    if os.path.join('openstackclient', 'tests', 'unit') not in filename:
        return

    if re.match(r'self.app = .*\(self.app, self.namespace\)', logical_line):
        yield (
            0,
            'O401: It is not necessary to create dummy Namespace objects',
        )

    if os.path.basename(filename) != 'fakes.py':
        if re.match(
            r'self.[a-zA-Z0-9_]+ = self.app.client_manager.*', logical_line
        ):
            yield (
                0,
                "O401: Aliases for mocks of the service client are already "
                "provided by the respective service's FakeClientMixin class",
            )

        if match := re.match(
            r'self.app.client_manager.([a-z_]+) = mock.Mock', logical_line
        ):
            service = match.group(1)
            if service == 'auth_ref':
                return
            yield (
                0,
                f"O401: client_manager.{service} mocks are already provided "
                f"by the {service} service's FakeClientMixin class",
            )


@core.flake8ext
def assert_use_of_client_aliases(logical_line, filename):
    """Ensure we use $service_client instead of $sdk_connection.service.

    Okay: self.compute_client.find_server(foo)
    O402: self.app.client_manager.sdk_connnection.compute.find_server(foo)

    The following checks only apply to files under openstackclient/tests/unit/:

    O402: self.app.client_manager.compute.find_server.return_value = server
    O402: self.app.client_manager.compute.find_server = mock.Mock()
    O402: self.compute_client.find_server = mock.Mock()
    """
    # we should expand the list of services as we drop legacy clients
    if match := re.match(
        r'self\.app\.client_manager\.sdk_connnection\.(compute|network|image)',
        logical_line,
    ):
        service = match.group(1)
        yield (0, f"O402: prefer {service}_client to sdk_connection.{service}")

    # everything from here down only affects unit tests
    if os.path.join('openstackclient', 'tests', 'unit') not in filename:
        return

    if match := re.match(
        r'(self\.app\.client_manager\.(compute|network|image)+\.[a-z_]+)\.(return_value|side_effect) = ',  # noqa: E501
        logical_line,
    ):
        service = match.group(1)
        yield (
            0,
            f"O402: prefer {service}_client to "
            f"self.app.client_manager.{service}",
        )

    if match := re.match(
        r'(self\.app\.client_manager\.(compute|network|image)+\.[a-z_]+) = mock.Mock',  # noqa: E501
        logical_line,
    ):
        yield (
            0,
            f"O402: {match.group(1)} is already a mock: there's no need to "
            f"assign a new mock.Mock instance.",
        )

    if match := re.match(
        r'(self\.(compute|network|image)_client\.[a-z_]+) = mock.Mock',
        logical_line,
    ):
        yield (
            0,
            f"O402: {match.group(1)} is already a mock: there's no need to "
            f"assign a new mock.Mock instance.",
        )


class SDKProxyFindChecker(ast.NodeVisitor):
    """NodeVisitor to find ``*_client.find_*`` statements."""

    def __init__(self):
        self.error = False

    def visit_Call(self, node):
        # No need to keep visiting the AST if we already found something.
        if self.error:
            return

        self.generic_visit(node)

        if not (
            isinstance(node.func, ast.Attribute)
            and node.func.attr.startswith('find_')  # and
            # isinstance(node.func.value, ast.Attribute) and
            # node.func.value.attr.endswith('_client')
        ):
            # print(f'skipping: got {node.func}')
            return

        if not (
            (
                # handle calls like 'identity_client.find_project'
                isinstance(node.func.value, ast.Name)
                and node.func.value.id.endswith('client')
            )
            or (
                # handle calls like 'self.compute_client.find_server'
                isinstance(node.func.value, ast.Attribute)
                and node.func.value.attr.endswith('_client')
            )
            or (
                # handle calls like 'self.app.client_manager.image.find_image'
                isinstance(node.func.value, ast.Attribute)
                and node.func.value.attr
                in ('identity', 'network', 'image', 'compute')
            )
        ):
            return

        if not any(kw.arg == 'ignore_missing' for kw in node.keywords):
            self.error = True


@core.flake8ext
def assert_find_ignore_missing_kwargs(logical_line, filename):
    """Ensure ignore_missing is always used for ``find_*`` SDK proxy calls.

    Okay: self.compute_client.find_server(foo, ignore_missing=True)
    Okay: self.image_client.find_server(foo, ignore_missing=False)
    Okay: self.volume_client.volumes.find(name='foo')
    O403: self.network_client.find_network(parsed_args.network)
    O403: self.compute_client.find_flavor(flavor_id, get_extra_specs=True)
    """
    if 'tests' in filename:
        return

    checker = SDKProxyFindChecker()
    try:
        parsed_logical_line = ast.parse(logical_line)
    except SyntaxError:
        # let flake8 catch this itself
        # https://github.com/PyCQA/flake8/issues/1948
        return
    checker.visit(parsed_logical_line)
    if checker.error:
        yield (
            0,
            'O403: Calls to find_* proxy methods must explicitly set '
            'ignore_missing',
        )


@core.flake8ext
def assert_use_of_osc_command(logical_line, filename):
    """Ensure we use openstackclient.command instead of osc_lib.command.

    Okay: from openstackclient.command import command
    Okay: import openstackclient.command
    O404: from osc_lib.command import command
    """
    if filename == 'openstackclient/command.py':
        return

    if re.match(r'^from osc_lib\.command import command$', logical_line):
        yield (
            0,
            'O404: Import Command classes from openstackclient.command, not '
            'osc_lib.command.command',
        )



#   Copyright 2015 iWeb Technologies Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""Volume v3 QoS action implementations"""

import argparse
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from openstack import utils as sdk_utils
from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


class AssociateQos(command.Command):
    _description = _("Associate a QoS specification to a volume type")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'qos_spec',
            metavar='<qos-spec>',
            help=_('QoS specification to modify (name or ID)'),
        )
        parser.add_argument(
            'volume_type',
            metavar='<volume-type>',
            help=_('Volume type to associate the QoS (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        qos_spec = volume_client.find_qos_spec(
            parsed_args.qos_spec, ignore_missing=False
        )
        volume_type = volume_client.find_type(
            parsed_args.volume_type, ignore_missing=False
        )
        volume_client.associate_qos_spec(qos_spec.id, volume_type.id)


class CreateQos(command.ShowOne):
    _description = _("Create new QoS specification")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help=_('New QoS specification name'),
        )
        consumer_choices = ['front-end', 'back-end', 'both']
        parser.add_argument(
            '--consumer',
            metavar='<consumer>',
            choices=consumer_choices,
            default='both',
            help=(
                _(
                    'Consumer of the QoS. Valid consumers: %s '
                    "(defaults to 'both')"
                )
                % utils.format_list(consumer_choices)
            ),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            dest='properties',
            action=parseractions.KeyValueAction,
            help=_(
                'Set a QoS specification property '
                '(repeat option to set multiple properties)'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        attrs: dict[str, Any] = {'consumer': parsed_args.consumer}
        if parsed_args.properties:
            attrs.update(parsed_args.properties)
        qos_spec = volume_client.create_qos_spec(
            name=parsed_args.name, **attrs
        )
        columns = ('consumer', 'id', 'name', 'properties')
        data = (
            qos_spec.consumer,
            qos_spec.id,
            qos_spec.name,
            format_columns.DictColumn(qos_spec.specs),
        )
        return columns, data


class DeleteQos(command.Command):
    _description = _("Delete QoS specification")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'qos_specs',
            metavar='<qos-spec>',
            nargs="+",
            help=_('QoS specification(s) to delete (name or ID)'),
        )
        parser.add_argument(
            '--force',
            action='store_true',
            default=False,
            help=_("Allow to delete in-use QoS specification(s)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        result = 0

        for i in parsed_args.qos_specs:
            try:
                qos_spec = volume_client.find_qos_spec(i, ignore_missing=False)
                volume_client.delete_qos_spec(
                    qos_spec.id, ignore_missing=False, force=parsed_args.force
                )
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete QoS specification with "
                        "name or ID '%(qos)s': %(e)s"
                    ),
                    {'qos': i, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.qos_specs)
            msg = _(
                "%(result)s of %(total)s QoS specifications failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class DisassociateQos(command.Command):
    _description = _("Disassociate a QoS specification from a volume type")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'qos_spec',
            metavar='<qos-spec>',
            help=_('QoS specification to modify (name or ID)'),
        )
        volume_type_group = parser.add_mutually_exclusive_group()
        volume_type_group.add_argument(
            '--volume-type',
            metavar='<volume-type>',
            help=_('Volume type to disassociate the QoS from (name or ID)'),
        )
        volume_type_group.add_argument(
            '--all',
            action='store_true',
            default=False,
            help=_('Disassociate the QoS from every volume type'),
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        qos_spec = volume_client.find_qos_spec(
            parsed_args.qos_spec, ignore_missing=False
        )

        if parsed_args.volume_type:
            volume_type = volume_client.find_type(
                parsed_args.volume_type, ignore_missing=False
            )
            volume_client.disassociate_qos_spec(qos_spec.id, volume_type.id)
        elif parsed_args.all:
            volume_client.disassociate_all_qos_spec(qos_spec.id)


class ListQos(command.Lister):
    _description = _("List QoS specifications")

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        qos_specs_list = list(volume_client.qos_specs())

        display_columns = (
            'ID',
            'Name',
            'Consumer',
            'Associations',
            'Properties',
        )

        data = []
        for qos in qos_specs_list:
            qos_associations = volume_client.qos_spec_associations(qos)
            associations = [a.name for a in qos_associations]
            data.append(
                (
                    qos.id,
                    qos.name,
                    qos.consumer,
                    format_columns.ListColumn(associations),
                    format_columns.DictColumn(qos.specs),
                )
            )

        return display_columns, iter(data)


class SetQos(command.Command):
    _description = _("Set QoS specification properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'qos_spec',
            metavar='<qos-spec>',
            help=_('QoS specification to modify (name or ID)'),
        )
        parser.add_argument(
            '--no-property',
            dest='no_property',
            action='store_true',
            help=_(
                'Remove all properties from <qos-spec> '
                '(specify both --no-property and --property to remove the '
                'current properties before setting new properties)'
            ),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            dest='properties',
            action=parseractions.KeyValueAction,
            help=_(
                'Property to add or modify for this QoS specification '
                '(repeat option to set multiple properties)'
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        qos_spec = volume_client.find_qos_spec(
            parsed_args.qos_spec, ignore_missing=False
        )

        result = 0
        if parsed_args.no_property:
            try:
                key_list = list(qos_spec.specs.keys())
                volume_client.delete_qos_spec_metadata(qos_spec.id, key_list)
            except Exception as e:
                LOG.error(_("Failed to clean qos properties: %s"), e)
                result += 1

        if parsed_args.properties:
            try:
                volume_client.update_qos_spec(
                    qos_spec.id, **parsed_args.properties
                )
            except Exception as e:
                LOG.error(_("Failed to set qos property: %s"), e)
                result += 1

        if result > 0:
            raise exceptions.CommandError(
                _("One or more of the set operations failed")
            )


class ShowQos(command.ShowOne):
    _description = _("Display QoS specification details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'qos_spec',
            metavar='<qos-spec>',
            help=_('QoS specification to display (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        qos_spec = volume_client.find_qos_spec(
            parsed_args.qos_spec, ignore_missing=False
        )
        qos_associations = list(volume_client.qos_spec_associations(qos_spec))
        associations = [a.name for a in qos_associations]
        if associations:
            columns: tuple[str, ...] = (
                'associations',
                'consumer',
                'id',
                'name',
                'properties',
            )
            data: tuple[Any, ...] = (
                format_columns.ListColumn(associations),
                qos_spec.consumer,
                qos_spec.id,
                qos_spec.name,
                format_columns.DictColumn(qos_spec.specs),
            )
        else:
            columns = ('consumer', 'id', 'name', 'properties')
            data = (
                qos_spec.consumer,
                qos_spec.id,
                qos_spec.name,
                format_columns.DictColumn(qos_spec.specs),
            )
        return columns, data


class UnsetQos(command.Command):
    _description = _("Unset QoS specification properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'qos_spec',
            metavar='<qos-spec>',
            help=_('QoS specification to modify (name or ID)'),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            action='append',
            dest='properties',
            default=[],
            help=_(
                'Property to remove from the QoS specification. '
                '(repeat option to unset multiple properties)'
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        qos_spec = volume_client.find_qos_spec(
            parsed_args.qos_spec, ignore_missing=False
        )
        if parsed_args.properties:
            volume_client.delete_qos_spec_metadata(
                qos_spec.id, parsed_args.properties
            )



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Volume v3 transfer action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack import exceptions as sdk_exceptions
from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


class AcceptTransferRequest(command.ShowOne):
    _description = _("Accept volume transfer request.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'transfer_request',
            metavar="<transfer-request-id>",
            help=_('Volume transfer request to accept (ID only)'),
        )
        parser.add_argument(
            '--auth-key',
            metavar="<key>",
            required=True,
            help=_('Volume transfer request authentication key'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        try:
            transfer_request_id = volume_client.find_transfer(
                parsed_args.transfer_request, ignore_missing=False
            ).id
        except sdk_exceptions.ResourceNotFound:
            # Non-admin users will fail to lookup name -> ID so we just
            # move on and attempt with the user-supplied information
            transfer_request_id = parsed_args.transfer_request

        result = volume_client.accept_transfer(
            transfer_request_id, parsed_args.auth_key
        )
        columns = ('id', 'name', 'volume_id')
        data = (result.id, result.name, result.volume_id)
        return columns, data


class CreateTransferRequest(command.ShowOne):
    _description = _("Create volume transfer request.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--name',
            metavar="<name>",
            help=_('New transfer request name (default to None)'),
        )
        parser.add_argument(
            '--snapshots',
            action='store_true',
            dest='snapshots',
            help=_(
                'Allow transfer volumes without snapshots (default) '
                '(supported by --os-volume-api-version 3.55 or later)'
            ),
            default=None,
        )
        parser.add_argument(
            '--no-snapshots',
            action='store_false',
            dest='snapshots',
            help=_(
                'Disallow transfer volumes without snapshots '
                '(supported by --os-volume-api-version 3.55 or later)'
            ),
        )
        parser.add_argument(
            'volume',
            metavar="<volume>",
            help=_('Volume to transfer (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        attrs: dict[str, Any] = {}

        if parsed_args.snapshots is not None:
            if not sdk_utils.supports_microversion(volume_client, '3.55'):
                msg = _(
                    "--os-volume-api-version 3.55 or greater is required to "
                    "support the '--(no-)snapshots' option"
                )
                raise exceptions.CommandError(msg)

            # unfortunately this option is negative so we have to reverse
            # things
            attrs['no_snapshots'] = not parsed_args.snapshots

        volume = volume_client.find_volume(
            parsed_args.volume, ignore_missing=False
        )
        result = volume_client.create_transfer(
            volume_id=volume.id,
            name=parsed_args.name,
            **attrs,
        )
        columns = ('auth_key', 'created_at', 'id', 'name', 'volume_id')
        data = (
            result.auth_key,
            result.created_at,
            result.id,
            result.name,
            result.volume_id,
        )
        return columns, data


class DeleteTransferRequest(command.Command):
    _description = _("Delete volume transfer request(s).")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'transfer_request',
            metavar="<transfer-request>",
            nargs="+",
            help=_('Volume transfer request(s) to delete (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        result = 0

        for t in parsed_args.transfer_request:
            try:
                volume_client.delete_transfer(t, ignore_missing=False)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete volume transfer request "
                        "with name or ID '%(transfer)s': %(e)s"
                    ),
                    {'transfer': t, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.transfer_request)
            msg = _(
                "%(result)s of %(total)s volume transfer requests failed"
                " to delete"
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListTransferRequest(command.Lister):
    _description = _("Lists all volume transfer requests.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--all-projects',
            dest='all_projects',
            action="store_true",
            default=False,
            help=_('Include all projects (admin only)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        column_headers = ('ID', 'Name', 'Volume')
        columns = ('id', 'name', 'volume_id')

        results = volume_client.transfers(
            details=True,
            all_projects=parsed_args.all_projects,
        )
        return (
            column_headers,
            (utils.get_item_properties(s, columns) for s in results),
        )


class ShowTransferRequest(command.ShowOne):
    _description = _("Show volume transfer request details.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'transfer_request',
            metavar="<transfer-request>",
            help=_('Volume transfer request to display (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        result = volume_client.find_transfer(
            parsed_args.transfer_request, ignore_missing=False
        )
        columns = ('created_at', 'id', 'name', 'volume_id')
        data = (result.created_at, result.id, result.name, result.volume_id)
        return columns, data



# Copyright (c) 2016, Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest.mock import call

from openstack.identity.v3 import domain as _domain
from openstack.identity.v3 import project as _project
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.network.v2 import network_meter_rule
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as tests_utils


class TestCreateMeterRule(network_fakes.TestNetworkV2):
    project = sdk_fakes.generate_fake_resource(_project.Project)
    domain = sdk_fakes.generate_fake_resource(_domain.Domain)
    new_rule = network_fakes.FakeNetworkMeterRule.create_one_rule()

    columns = (
        'destination_ip_prefix',
        'direction',
        'excluded',
        'id',
        'metering_label_id',
        'project_id',
        'remote_ip_prefix',
        'source_ip_prefix',
    )
    data = (
        new_rule.destination_ip_prefix,
        new_rule.direction,
        new_rule.excluded,
        new_rule.id,
        new_rule.metering_label_id,
        new_rule.project_id,
        new_rule.remote_ip_prefix,
        new_rule.source_ip_prefix,
    )

    def setUp(self):
        super().setUp()
        fake_meter = network_fakes.FakeNetworkMeter.create_one_meter(
            {'id': self.new_rule.metering_label_id}
        )
        self.network_client.create_metering_label_rule.return_value = (
            self.new_rule
        )
        self.identity_sdk_client.find_project.return_value = self.project
        self.network_client.find_metering_label.return_value = fake_meter

        self.cmd = network_meter_rule.CreateMeterRule(self.app, None)

    def test_create_no_options(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_default_options(self):
        arglist = [
            self.new_rule.metering_label_id,
            "--remote-ip-prefix",
            self.new_rule.remote_ip_prefix,
        ]
        verifylist = [
            ('meter', self.new_rule.metering_label_id),
            ('remote_ip_prefix', self.new_rule.remote_ip_prefix),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_metering_label_rule.assert_called_once_with(
            **{
                'direction': 'ingress',
                'metering_label_id': self.new_rule.metering_label_id,
                'remote_ip_prefix': self.new_rule.remote_ip_prefix,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_all_options(self):
        arglist = [
            "--ingress",
            "--include",
            self.new_rule.metering_label_id,
            "--remote-ip-prefix",
            self.new_rule.remote_ip_prefix,
        ]
        verifylist = [
            ('ingress', True),
            ('include', True),
            ('meter', self.new_rule.metering_label_id),
            ('remote_ip_prefix', self.new_rule.remote_ip_prefix),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_metering_label_rule.assert_called_once_with(
            **{
                'direction': self.new_rule.direction,
                'excluded': self.new_rule.excluded,
                'metering_label_id': self.new_rule.metering_label_id,
                'remote_ip_prefix': self.new_rule.remote_ip_prefix,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


class TestDeleteMeterRule(network_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()
        self.rule_list = network_fakes.FakeNetworkMeterRule.create_meter_rule(
            count=2
        )
        self.network_client.delete_metering_label_rule.return_value = None

        self.network_client.find_metering_label_rule = (
            network_fakes.FakeNetworkMeterRule.get_meter_rule(
                meter_rule=self.rule_list
            )
        )

        self.cmd = network_meter_rule.DeleteMeterRule(self.app, None)

    def test_delete_one_rule(self):
        arglist = [
            self.rule_list[0].id,
        ]
        verifylist = [
            ('meter_rule_id', [self.rule_list[0].id]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_metering_label_rule.assert_called_once_with(
            self.rule_list[0]
        )
        self.assertIsNone(result)

    def test_delete_multiple_rules(self):
        arglist = []
        for rule in self.rule_list:
            arglist.append(rule.id)
        verifylist = [
            ('meter_rule_id', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        calls = []
        for rule in self.rule_list:
            calls.append(call(rule))
        self.network_client.delete_metering_label_rule.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_delete_multiple_rules_exception(self):
        arglist = [
            self.rule_list[0].id,
            'xxxx-yyyy-zzzz',
            self.rule_list[1].id,
        ]
        verifylist = [
            ('meter_rule_id', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        return_find = [
            self.rule_list[0],
            exceptions.NotFound('404'),
            self.rule_list[1],
        ]
        self.network_client.find_metering_label_rule.side_effect = return_find

        ret_delete = [
            None,
            exceptions.NotFound('404'),
        ]
        self.network_client.delete_metering_label_rule.side_effect = ret_delete

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

        calls = [
            call(self.rule_list[0]),
            call(self.rule_list[1]),
        ]
        self.network_client.delete_metering_label_rule.assert_has_calls(calls)


class TestListMeterRule(network_fakes.TestNetworkV2):
    rule_list = network_fakes.FakeNetworkMeterRule.create_meter_rule(count=2)

    columns = (
        'ID',
        'Excluded',
        'Direction',
        'Remote IP Prefix',
        'Source IP Prefix',
        'Destination IP Prefix',
    )

    data = []

    for rule in rule_list:
        data.append(
            (
                rule.id,
                rule.excluded,
                rule.direction,
                rule.remote_ip_prefix,
                rule.source_ip_prefix,
                rule.destination_ip_prefix,
            )
        )

    def setUp(self):
        super().setUp()

        self.network_client.metering_label_rules.return_value = self.rule_list

        self.cmd = network_meter_rule.ListMeterRule(self.app, None)

    def test_rule_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.metering_label_rules.assert_called_with()
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))


class TestShowMeterRule(network_fakes.TestNetworkV2):
    new_rule = network_fakes.FakeNetworkMeterRule.create_one_rule()

    columns = (
        'destination_ip_prefix',
        'direction',
        'excluded',
        'id',
        'metering_label_id',
        'project_id',
        'remote_ip_prefix',
        'source_ip_prefix',
    )

    data = (
        new_rule.destination_ip_prefix,
        new_rule.direction,
        new_rule.excluded,
        new_rule.id,
        new_rule.metering_label_id,
        new_rule.project_id,
        new_rule.remote_ip_prefix,
        new_rule.source_ip_prefix,
    )

    def setUp(self):
        super().setUp()

        self.cmd = network_meter_rule.ShowMeterRule(self.app, None)

        self.network_client.find_metering_label_rule.return_value = (
            self.new_rule
        )

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_label_rule_show_option(self):
        arglist = [
            self.new_rule.id,
        ]
        verifylist = [
            ('meter_rule_id', self.new_rule.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.find_metering_label_rule.assert_called_with(
            self.new_rule.id, ignore_missing=False
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Meter Rule Implementations"""

import argparse
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from openstack.network.v2 import metering_label_rule as _metering_label_rule
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common
from openstackclient.network import common

LOG = logging.getLogger(__name__)


def _get_columns(
    item: _metering_label_rule.MeteringLabelRule,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    hidden_columns = ['location', 'tenant_id']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, {}, hidden_columns
    )


def _get_attrs(
    client_manager: Any, parsed_args: argparse.Namespace
) -> dict[str, object]:
    attrs: dict[str, object] = {}

    if parsed_args.exclude:
        attrs['excluded'] = True
    if parsed_args.include:
        attrs['excluded'] = False
    if parsed_args.ingress or not parsed_args.egress:
        attrs['direction'] = 'ingress'
    if parsed_args.egress:
        attrs['direction'] = 'egress'
    if parsed_args.remote_ip_prefix is not None:
        attrs['remote_ip_prefix'] = parsed_args.remote_ip_prefix
    if parsed_args.source_ip_prefix is not None:
        attrs['source_ip_prefix'] = parsed_args.source_ip_prefix
    if parsed_args.destination_ip_prefix is not None:
        attrs['destination_ip_prefix'] = parsed_args.destination_ip_prefix
    if parsed_args.meter is not None:
        attrs['metering_label_id'] = parsed_args.meter
    if parsed_args.project is not None:
        identity_client = client_manager.sdk_connection.identity
        project_id = identity_common.find_project_id_sdk(
            identity_client,
            parsed_args.project,
            parsed_args.project_domain,
        )
        attrs['project_id'] = project_id

    return attrs


class CreateMeterRule(command.ShowOne, common.NeutronCommandWithExtraArgs):
    _description = _("Create a new meter rule")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)

        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_("Owner's project (name or ID)"),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        exclude_group = parser.add_mutually_exclusive_group()
        exclude_group.add_argument(
            '--exclude',
            action='store_true',
            help=_("Exclude remote IP prefix from traffic count"),
        )
        exclude_group.add_argument(
            '--include',
            action='store_true',
            help=_("Include remote IP prefix from traffic count (default)"),
        )
        direction_group = parser.add_mutually_exclusive_group()
        direction_group.add_argument(
            '--ingress',
            action='store_true',
            help=_("Apply rule to incoming network traffic (default)"),
        )
        direction_group.add_argument(
            '--egress',
            action='store_true',
            help=_('Apply rule to outgoing network traffic'),
        )
        parser.add_argument(
            '--remote-ip-prefix',
            metavar='<remote-ip-prefix>',
            required=False,
            help=_('The remote IP prefix to associate with this rule'),
        )
        parser.add_argument(
            '--source-ip-prefix',
            metavar='<remote-ip-prefix>',
            required=False,
            help=_('The source IP prefix to associate with this rule'),
        )
        parser.add_argument(
            '--destination-ip-prefix',
            metavar='<remote-ip-prefix>',
            required=False,
            help=_('The destination IP prefix to associate with this rule'),
        )
        parser.add_argument(
            'meter',
            metavar='<meter>',
            help=_('Label to associate with this metering rule (name or ID)'),
        )

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        _meter = client.find_metering_label(
            parsed_args.meter, ignore_missing=False
        )
        parsed_args.meter = _meter.id
        attrs = _get_attrs(self.app.client_manager, parsed_args)
        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )
        obj = client.create_metering_label_rule(**attrs)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns, formatters={})

        return (display_columns, data)


class DeleteMeterRule(command.Command):
    _description = _("Delete meter rule(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'meter_rule_id',
            metavar='<meter-rule-id>',
            nargs='+',
            help=_('Meter rule to delete (ID only)'),
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        result = 0

        for id in parsed_args.meter_rule_id:
            try:
                obj = client.find_metering_label_rule(id, ignore_missing=False)
                client.delete_metering_label_rule(obj)
            except Exception as e:
                result += 1
                LOG.error(
                    _("Failed to delete meter rule with ID '%(id)s': %(e)s"),
                    {"id": id, "e": e},
                )

        if result > 0:
            total = len(parsed_args.meter_rule_id)
            msg = _(
                "%(result)s of %(total)s meter rules failed to delete."
            ) % {"result": result, "total": total}
            raise exceptions.CommandError(msg)


class ListMeterRule(command.Lister):
    _description = _("List meter rules")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        pagination.add_marker_pagination_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network

        columns = (
            'id',
            'excluded',
            'direction',
            'remote_ip_prefix',
            'source_ip_prefix',
            'destination_ip_prefix',
        )
        column_headers = (
            'ID',
            'Excluded',
            'Direction',
            'Remote IP Prefix',
            'Source IP Prefix',
            'Destination IP Prefix',
        )

        filters = {}
        if parsed_args.marker is not None:
            filters['marker'] = parsed_args.marker
        if parsed_args.limit is not None:
            filters['limit'] = parsed_args.limit
        if parsed_args.max_items is not None:
            filters['max_items'] = parsed_args.max_items

        data = client.metering_label_rules(**filters)
        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                )
                for s in data
            ),
        )


class ShowMeterRule(command.ShowOne):
    _description = _("Display meter rules details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'meter_rule_id',
            metavar='<meter-rule-id>',
            help=_('Meter rule (ID only)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        obj = client.find_metering_label_rule(
            parsed_args.meter_rule_id, ignore_missing=False
        )
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns)
        return display_columns, data



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from typing import Any, cast

from osc_lib import exceptions

from openstackclient.i18n import _


def str2bool(strbool: str | None) -> bool | None:
    if strbool is None:
        return None
    return strbool.lower() == 'true'


def str2list(strlist: str | None) -> list[str]:
    result: list[str] = []
    if strlist:
        result = strlist.split(';')
    return result


def str2dict(strdict: str) -> dict[str, str]:
    """Convert key1:value1;key2:value2;... string into dictionary.

    :param strdict: string in the form of key1:value1;key2:value2
    """
    result: dict[str, str] = {}
    if not strdict:
        return result
    i = 0
    kvlist = []
    for kv in strdict.split(';'):
        if ':' in kv:
            kvlist.append(kv)
            i += 1
        elif i == 0:
            msg = _("missing value for key '%s'")
            raise exceptions.CommandError(msg % kv)
        else:
            kvlist[i - 1] = f"{kvlist[i - 1]};{kv}"
    for kv in kvlist:
        key, value = kv.split(':', 1)
        result[key] = value
    return result


def format_network_port_range(rule: dict[str, Any]) -> str:
    # Display port range or ICMP type and code. For example:
    # - ICMP type: 'type=3'
    # - ICMP type and code: 'type=3:code=0'
    # - ICMP code: Not supported
    # - Matching port range: '443:443'
    # - Different port range: '22:24'
    # - Single port: '80:80'
    # - No port range: ''
    port_range = ''
    if is_icmp_protocol(rule['protocol']):
        if rule['port_range_min']:
            port_range += 'type=' + str(rule['port_range_min'])
        if rule['port_range_max']:
            port_range += ':code=' + str(rule['port_range_max'])
    elif rule['port_range_min'] or rule['port_range_max']:
        port_range_min = str(rule['port_range_min'])
        port_range_max = str(rule['port_range_max'])
        if rule['port_range_min'] is None:
            port_range_min = port_range_max
        if rule['port_range_max'] is None:
            port_range_max = port_range_min
        port_range = port_range_min + ':' + port_range_max
    return port_range


def format_remote_ip_prefix(rule: dict[str, Any]) -> str | None:
    remote_ip_prefix = cast(str | None, rule['remote_ip_prefix'])
    if remote_ip_prefix is None:
        ethertype = rule['ether_type']
        if ethertype == 'IPv4':
            remote_ip_prefix = '0.0.0.0/0'
        elif ethertype == 'IPv6':
            remote_ip_prefix = '::/0'
    return remote_ip_prefix


def convert_ipvx_case(string: str) -> str:
    if string.lower() == 'ipv4':
        return 'IPv4'
    if string.lower() == 'ipv6':
        return 'IPv6'
    return string


def is_icmp_protocol(protocol: str | None) -> bool:
    # NOTE(rtheis): Neutron has deprecated protocol icmpv6.
    # However, while the OSC CLI doesn't document the protocol,
    # the code must still handle it. In addition, handle both
    # protocol names and numbers.
    if protocol in ['icmp', 'icmpv6', 'ipv6-icmp', '1', '58']:
        return True
    else:
        return False


def convert_to_lowercase(string: str) -> str:
    return string.lower()


def get_protocol(
    parsed_args: Any, default_protocol: str = 'any'
) -> str | None:
    protocol: str | None = default_protocol
    if parsed_args.protocol is not None:
        protocol = parsed_args.protocol
    if hasattr(parsed_args, "proto") and parsed_args.proto is not None:
        protocol = parsed_args.proto
    if protocol == 'any':
        protocol = None
    return protocol


def get_ethertype(parsed_args: Any, protocol: str | None) -> str:
    ethertype = 'IPv4'
    if parsed_args.ethertype is not None:
        ethertype = parsed_args.ethertype
    elif is_ipv6_protocol(protocol):
        ethertype = 'IPv6'
    return ethertype


def is_ipv6_protocol(protocol: str | None) -> bool:
    # NOTE(rtheis): Neutron has deprecated protocol icmpv6.
    # However, while the OSC CLI doesn't document the protocol,
    # the code must still handle it. In addition, handle both
    # protocol names and numbers.
    if (protocol is not None and protocol.startswith('ipv6-')) or protocol in [
        'icmpv6',
        '41',
        '43',
        '44',
        '58',
        '59',
        '60',
    ]:
        return True
    else:
        return False



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import copy

from keystoneauth1 import token_endpoint
from osc_lib.tests import utils as osc_lib_test_utils

from openstackclient.common import clientmanager
from openstackclient.tests.unit import fakes


class TestClientManager(osc_lib_test_utils.TestClientManager):
    def _clientmanager_class(self):
        """Allow subclasses to override the ClientManager class"""
        return clientmanager.ClientManager

    def test_client_manager_admin_token(self):
        token_auth = {
            'endpoint': fakes.AUTH_URL,
            'token': fakes.AUTH_TOKEN,
        }
        client_manager = self._make_clientmanager(
            auth_args=token_auth,
            auth_plugin_name='admin_token',
        )

        self.assertEqual(
            fakes.AUTH_URL,
            client_manager._cli_options.config['auth']['endpoint'],
        )
        self.assertEqual(
            fakes.AUTH_TOKEN,
            client_manager.auth.get_token(None),
        )
        self.assertIsInstance(
            client_manager.auth,
            token_endpoint.Token,
        )
        self.assertTrue(client_manager.is_network_endpoint_enabled())

    def test_client_manager_network_endpoint_disabled(self):
        auth_args = copy.deepcopy(self.default_password_auth)
        auth_args.update(
            {
                'user_domain_name': 'default',
                'project_domain_name': 'default',
            }
        )
        # v3 fake doesn't have network endpoint
        client_manager = self._make_clientmanager(
            auth_args=auth_args,
            identity_api_version='3',
            auth_plugin_name='v3password',
        )

        self.assertFalse(client_manager.is_service_available('network'))
        # This is True because ClientManager.auth_ref returns None in this
        # test; "no service catalog" means use Network API by default now
        self.assertTrue(client_manager.is_network_endpoint_enabled())
